/*
	This is a ACLfs client.

	It maps various filesystems into the client home/aclfs directory.
	It deals with the aclfsd server to gain access to files. The later
	implement (may implement) sophisticated access control using
	ACL (Access Control List) instead of the limited Unix 9 bits
	permission scheme.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <utime.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "aclfs.h"
#include "remaudio.h"
#include "dispatch.h"

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
#else
	typedef struct ucred ACL_UCRED;
#endif

extern "C" char *stpcpy (char *, const char *);

struct ACLFS_CONF {
	char *volume;
	char *realvolume;
};

static int nbdef=0;
static ACLFS_CONF *conf;

static void aclfs_readconf ()
{
	if (conf == NULL){
		/* #Specification: .aclfs / global and user config
			The aclfs configuration for a user is the union of
			/etc/aclfs.conf and the user's .aclfs file. If one
			aclfs volume exist in both /etc/aclfs.conf and the
			.aclfs file, the user definition has priority.

			Note that this behavior is weak. We must implement
			something better that merge all entries in the "n"
			directory. A user may have a logical volume in
			litefs that conflict with /etc/aclfs.conf, and the
			definition in .litefs must have priority.

			This will be solve one day once we have a virtualfs
			agent that loads all config file and make the configuration
			available through a shared memory mapping. This utility
			will be able to handle the .persmount, .litefs and .aclfs
			(and what's next :-) ) properly.
		*/
		conf = (ACLFS_CONF*)malloc (100*sizeof(ACLFS_CONF));
		FILE *fin = fopen ("/etc/aclfs.conf","r");
		if (fin != NULL){
			char buf[1000];
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char *pt = buf;
				while (isspace (*pt)) pt++;
				if (pt[0] != '\0' && pt[0] != '#'){
					char *volume = pt;
					while (*pt > ' ') pt++;
					if (isspace(*pt)) *pt = '\0';
					conf[nbdef].volume = strdup(volume);
					conf[nbdef].realvolume = strdup(volume);
					nbdef++;
				}
			}
			fclose (fin);
		}
		char path[PATH_MAX];
		snprintf (path,sizeof(path)-1,"%s/.aclfs",_dispatch_gethome());
		fin = fopen (path,"r");
		if (fin != NULL){
			char buf[1000];
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char *pt = buf;
				while (isspace (*pt)) pt++;
				if (pt[0] != '\0' && pt[0] != '#'){
					char *volume = pt;
					while (*pt > ' ') pt++;
					if (isspace(*pt)){
						*pt ++ = '\0';
						while (isspace(*pt)) pt++;
						char *end = pt + strlen(pt)-1;
						if (*end == '\n') *end = '\0';
						int i;
						for (i=0; i<nbdef; i++){
							if (strcmp(conf[i].volume,volume)==0){
								free (conf[i].realvolume);
								conf[i].realvolume = strdup(pt);
								break;
							}
						}
						if (i==nbdef){
							conf[nbdef].volume = strdup(volume);
							conf[nbdef].realvolume = strdup(pt);
							nbdef++;
						}
					}
				}
			}
			fclose (fin);
		}
	}
}

struct ACLFS_INFO{
	char volume[PATH_MAX];		// Sub-directory in $HOME/n
	char relpath[PATH_MAX];		// relative path in $HOME/n/volume/
	char virtpath[PATH_MAX];	// absolute virtual path
	char realpath[PATH_MAX];	// Translated path
};

/*
	Check if a volume is configured for the personal auto mounter
*/
static int aclfs_chkconf (ACLFS_INFO &info)
{
	aclfs_readconf();
	int ret = -1;
	for (int i=0; i<nbdef; i++){
		if (strcmp(info.volume,conf[i].volume)==0){
			sprintf (info.realpath,"/%s%s",conf[i].realvolume
				,info.relpath);
			ret = 0;
			break;
		}
	}
	return ret;
}

static int acl_fd=-1;	// Socket to do request

/*
	Wait to receive a file handle
	Return the file handle or -1.
*/
static int aclfs_receivefd()
{
	int ret = -1;
	char bufctl[sizeof(struct cmsghdr)+1*sizeof(int)];
	memset (bufctl,-1,sizeof(bufctl));
	struct iovec iov[1];
	char buf[1000];
	iov[0].iov_base = buf;
	iov[0].iov_len = sizeof(buf);
	struct msghdr h;
	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = bufctl;
	h.msg_controllen = sizeof(bufctl);
	struct cmsghdr *c = (struct cmsghdr*) bufctl;
	c->cmsg_len = sizeof(bufctl);
	//c->cmsg_level = SOL_SOCKET;
	//c->cmsg_type = SCM_RIGHTS;
	h.msg_flags = 0;
	int len = recvmsg (acl_fd,&h,0);
	if (len<=0){
		perror ("recvmsg");
	}else{
		#if 0
			fprintf (stderr,"message recu flags %d len %d %d %u\n"
				,h.msg_flags,len
				,h.msg_controllen,sizeof(bufctl));
			fprintf (stderr,"iov.len %d\n",iov[0].iov_len);
			buf[len] = '\0';
			fprintf (stderr,"buf = :%s:\n",buf);
			fprintf (stderr,"msg_len %d msg_level %d msg_type %d\n"
				,c->cmsg_len
				,c->cmsg_level
				,c->cmsg_type);
		#endif
		/*
			On kernel 2.2, we can check c->cmsg_type and c->cmsg_level
			to find out if there is a file handle

			On kernel 2.0, we can only tell if there is
			a descriptor if it is != -1
			if (c->cmsg_level == SOL_SOCKET
				&& c->cmsg_type == SCM_RIGHTS){
		*/
		int *fd = (int*)CMSG_DATA(c);
		//fprintf (stderr,"fd passing %d\n",fd[0]);
		ret = fd[0];
	}
	return ret;
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
static int aclfs_receive()
{
	int ret = -1;
	union {
		char buf[100];
		ACL_REPLY rep;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	if (len == sizeof(ACL_REPLY)){
		if (u.rep.status == ACL_STATUS_OK){
			ret = 0;
		}else{
			errno = u.rep.errcode;
		}
	}
	return ret;
}
/*
	Wait for some data
	Return -1 if the original request failed.
*/
static int aclfs_receive_data(char *buf, int size)
{
	int ret = -1;
	ACL_REPLY_DATA rep;
	int len = read (acl_fd,&rep,sizeof(rep));
	if (len >= (int)sizeof(ACL_REPLY)){
		if (rep.status == ACL_STATUS_DATA){
			if (size > rep.len) size = rep.len;
			memcpy (buf,rep.data,size);
			ret = size;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}

/*
	Wait for a confirmation of a stat request.
	Return -1 if the original request failed.
*/
static int aclfs_receive_stat (struct stat &buf)
{
	int ret = -1;
	struct {
		ACL_REPLY rep;
		struct stat st;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	memset (&buf,0,sizeof(struct stat));
	// fprintf (stderr,"receive_stat %d %d %d\n",len,sizeof(u),u.rep.status);
	if (len >= (int)sizeof(ACL_REPLY)){
		errno = u.rep.errcode;
		if (u.rep.status == ACL_STATUS_STAT
			&& u.rep.len == sizeof(struct stat)){
			memcpy (&buf,&u.st,sizeof(struct stat));
			ret = 0;
		}
	}
	return ret;
}

static int aclfs_connect ()
{
	int ret = -1;
	if (acl_fd != -1){
		ret = 0;
	}else{
		int fd =  socket (AF_UNIX,SOCK_STREAM,0);
		if (fd == -1){
			perror("socket client");
		}else{
			struct sockaddr_un un;
			un.sun_family = AF_UNIX;
			strcpy (un.sun_path,ACLFS_SOCKET);
			int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
			if (s == -1){
				perror("connect");
			}else{
				acl_fd = fd;
				ret = 0;
			}
		}
	}
	return ret;
}

/*
	Translate a logical path if it is part of the personnal mount area
	($HOME/n)
*/
static int aclfs_translate (
	const char *path,
	ACLFS_INFO &info,
	ACL_REQUEST &req,
	int &len_request)
{
	int lenn;
	const char *pathn = _dispatch_getpathn(lenn);
	if (pathn == NULL) return -1;
	strcpy (info.realpath,path);
	if (path[0] != '/'){
		const char *curpath = _dispatch_getcwd();
		snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
		path = info.virtpath;
	}else{
		strcpy (info.virtpath,path);
	}
	int ret = -1;
	if (strstr(path,"..")!=NULL){
		_dispatch_simplifypath (info.virtpath,info.virtpath);
		path = info.virtpath;
	}

	//printf ("pathn = :%s: <> :%s:\n",pathn,path);
	if (strncmp(pathn,path,lenn)==0){
		path += lenn;
		char *volume = info.volume;
		while (*path != '\0' && *path != '/'){
			*volume++ = *path++;
		}
		*volume = '\0';
		strcpy (info.relpath,path);
		//printf ("volume :%s: :%s:\n",info.volume,info.relpath);
		if (aclfs_chkconf (info)!=-1){
			if (aclfs_connect() != -1){
				char *end = stpcpy (req.paths,info.realpath) + 1;
				len_request = (int)(end-(char*)&req);
				req.mode = 0;
				req.perm = 0;
				ret = 0;
			}
		}
	}
	return ret;
}

static void aclfs_sendreq (ACL_REQUEST &req, int len)
{
	char bufctl[sizeof(struct cmsghdr)+sizeof(ACL_UCRED)];
	struct cmsghdr *c = (struct cmsghdr*)bufctl;
	c->cmsg_len = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_CREDENTIALS;
	ACL_UCRED *uc = (ACL_UCRED*)CMSG_DATA(c);
	uc->pid = getpid();
	uc->uid = getuid();
	uc->gid = getgid();


	struct msghdr h;

	struct iovec iov[1];
	iov[0].iov_base = &req;
	iov[0].iov_len = len;

	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = (void*)bufctl;
	h.msg_controllen = sizeof(bufctl);
	h.msg_flags = 0;
	void (*old)(int) = signal (SIGPIPE,SIG_IGN);
	int ret = sendmsg (acl_fd,&h,0);
	if (ret==-1){
		for (int i=0; i<3; i++){
			acl_fd = -1;
			aclfs_connect();
			ret = sendmsg (acl_fd,&h,0);
			if (ret == -1 && i!=2) sleep(1);
		}
	}
	if (ret == -1){		
		perror ("sendmsg");
	}
	signal (SIGPIPE,old);
}


static int aclfs_open (const char *path, int mode, int perm)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		//fprintf (stderr,"open :%s: %d %d\n",info.realpath,mode,perm);
		req.cmd = ACL_OPEN;
		req.mode = mode;
		req.perm = perm;
		aclfs_sendreq (req,len);
		ret = aclfs_receivefd();
	}
	return ret;
}

static int aclfs_mkdiranc (const char *path)
{
	int ret = __mkdir (path,0755);
	if (ret == -1){
		char tmp[strlen(path)+1];
		strcpy (tmp,path);
		char *pt = strrchr(tmp,'/');
		if (pt != NULL){
			*pt = '\0';
			if (aclfs_mkdiranc (tmp) != -1) ret = __mkdir (path,0755);
		}
	}
	return ret;
}


static int aclfs_chdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		//fprintf (stderr,"chdir :%s: :%s: :%s:\n",path,info.virtpath,info.realpath);
		req.cmd = ACL_CHDIR;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		if (ret != -1){
			// We have to set the new current dir in $HOME/n
			if (__chdir (info.virtpath) == -1){
				aclfs_mkdiranc (info.virtpath);
				__chdir (info.virtpath);
			}
		}
	}
	return ret;
}

static struct DIR * aclfs_opendir (const char *path)
{
	struct DIR *ret = NULL;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_OPENDIR;
		aclfs_sendreq (req,len);
		int fd = aclfs_receivefd();
		if (fd != -1){
			ret = (DIR*)fd;
		}
	}
	return ret;
}

static struct dirent *aclfs_readdir (struct DIR *dir)
{
	int fd = (int)dir;
	static struct dirent ent;
	struct dirent *ret = NULL;
	if (read (fd,&ent,sizeof(ent))==sizeof(ent)) ret = &ent;
	return ret;
}


static int aclfs_closedir (struct DIR *dir)
{
	int fd = (int)dir;
	return close (fd);
}
static int aclfs_rename (const char *oldp, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO oldinfo,newinfo;
	ACL_REQUEST oldreq,newreq;
	int oldlen,newlen;
	int oldf = aclfs_translate (oldp,oldinfo,oldreq,oldlen);
	int newf = aclfs_translate (newp,newinfo,newreq,newlen);
	// fprintf (stderr,"aclfs_rename %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (oldf != -1 && newf !=-1
		&& strcmp(oldinfo.volume,newinfo.volume)==0){
		// fprintf (stderr,"aclfs_rename :%s:[%d] :%s:[%d]\n"
		//	,oldreq.paths,oldlen
		//	,newreq.paths,newlen);
		oldreq.cmd = ACL_RENAME;
		strcpy (oldreq.paths+strlen(oldreq.paths)+1,newreq.paths);
		aclfs_sendreq (oldreq,oldlen+strlen(newreq.paths)+1);
		ret = aclfs_receive();
	}else if (oldf != -1 || newf != -1){
		errno = EXDEV;
		ret = -1;
	}
	return ret;
}
static int aclfs_unlink (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"unlink :%s:\n",info.realpath);
		req.cmd = ACL_UNLINK;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_rmdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_RMDIR;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_mkdir(const char *path, int mode)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_MKDIR;
		req.mode = mode;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

static int aclfs_stat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"stat :%s:\n",info.realpath);
		req.cmd = ACL_STAT;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_stat(*buf);
	}
	return ret;
}
static int aclfs_lstat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"lstat :%s:\n",info.realpath);
		req.cmd = ACL_LSTAT;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_stat(*buf);
	}
	return ret;
}

static int aclfs_chmod (const char *fname, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_CHMOD;
		req.mode = mode;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_chown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_CHOWN;
		req.mode = uid;
		req.perm = gid;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_lchown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_LCHOWN;
		req.mode = uid;
		req.perm = gid;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}


static int aclfs_readlink (const char *fname, char *buf, size_t size)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_READLINK;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_data(buf,size);
	}
	return ret;
}
#if 0
static int aclfs_utimes (const char *fname, struct timeval *tvp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_UTIMES;
		if (tvp == NULL){
			req.mode = -1;
			req.perm = -1;
		}else{
			req.mode = tvp[0].tv_sec;
			req.perm = tvp[1].tv_sec;
		}
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
#endif
static int aclfs_utime (const char *fname, const struct utimbuf *tvp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_UTIMES;
		if (tvp == NULL){
			req.mode = -1;
			req.perm = -1;
		}else{
			req.mode = tvp->actime;
			req.perm = tvp->modtime;
		}
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

static int aclfs_access (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_ACCESS;
		req.mode = type;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		//fprintf (stderr,"access :%s: ret=%d errno=%d\n",fname,ret,errno);
	}
	return ret;
}

static int aclfs_euidaccess (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_EUIDACCESS;
		req.mode = type;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		//fprintf (stderr,"euidaccess :%s: ret=%d errno=%d\n",fname,ret,errno);
	}
	return ret;
}

static int aclfs_link (const char *oldp, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO oldinfo,newinfo;
	ACL_REQUEST oldreq,newreq;
	int oldlen,newlen;
	int oldf = aclfs_translate (oldp,oldinfo,oldreq,oldlen);
	int newf = aclfs_translate (newp,newinfo,newreq,newlen);
	// fprintf (stderr,"aclfs_link %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (oldf != -1 && newf !=-1
		&& strcmp(oldinfo.volume,newinfo.volume)==0){
		oldreq.cmd = ACL_LINK;
		strcpy (oldreq.paths+oldlen,newinfo.relpath);
		aclfs_sendreq (oldreq,oldlen+newlen);
		ret = aclfs_receive();
	}else if (oldf != -1 || newf != -1){
		errno = EXDEV;
		ret = -1;
	}
	return ret;
}
static int aclfs_symlink (const char *oldp, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO newinfo;
	ACL_REQUEST newreq;
	int newlen;
	// fprintf (stderr,"aclfs_symlink %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (aclfs_translate (newp,newinfo,newreq,newlen)!=-1){
		// oldp does not have to lie in the acl filesystem and do
		// not requires validation on our side. To make things
		// simpler on the remote side, we pass the argument
		// using the same order as rename and link.
		newreq.cmd = ACL_SYMLINK;
		char tmp[PATH_MAX];
		strncpy (tmp,newreq.paths,PATH_MAX);
		strcpy (newreq.paths,oldp);
		int oldlen = strlen(oldp)+1;
		strcpy (newreq.paths+oldlen,tmp);
		aclfs_sendreq (newreq,newlen+oldlen);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_fchmod (int, mode_t)
{
	errno = EPERM;
	return -1;
}
static int aclfs_fchown (int, uid_t, gid_t)
{
	errno = EPERM;
	return -1;
}

static int aclfs_execve (const char *filename, char *const[], char *const[])
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (filename,info,req,len)!=-1){
		ret = -1;
		errno = EPERM;
	}
	return ret;
}



static DISPATCH m={
	NULL,
	aclfs_open,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	aclfs_stat,
	NULL,
	aclfs_lstat,
	aclfs_chdir,
	aclfs_opendir,
	aclfs_readdir,
	aclfs_closedir,
	aclfs_rename,
	aclfs_unlink,
	aclfs_rmdir,
	aclfs_mkdir,
	NULL,		// socket
	NULL,		// connect
	NULL,		// bind
	aclfs_chmod,
	aclfs_chown,
	aclfs_lchown,
	aclfs_readlink,
	aclfs_utime,
	aclfs_access,
	aclfs_euidaccess,
	aclfs_link,
	aclfs_symlink,
	aclfs_fchmod,
	aclfs_fchown,
	aclfs_execve,
};

static REGISTER_DISPATCH register_m (m);


