#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <syslog.h>
#include "aclfs.h"
#include "aclfsd.h"

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
	#define SO_PASSCRED   16 
#else
	typedef struct ucred ACL_UCRED;
#endif

static bool debug=false;

void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}

static void usage()
{
	fprintf (stderr,"aclfsd version 1.0\n"
		"\n"
		"aclfsd [ --debug ] [ --nodaemon ] [ --socket path ] [ --pidfile path ]\n"
		"       [ --config path ]\n"
		"aclfsd --testacl dir file user group\n"
		"aclfsd --reload dir\n"
		"aclfsd --reloadconf\n"
		"aclfsd --help\n");
}


static int aclfsd_server(const char *sockn)
{
	int ret = -1;
	unlink (sockn);
	int fd =  socket (AF_UNIX,SOCK_STREAM,0);
	if (fd == -1){
		perror("socket server");
	}else{
		struct sockaddr_un un;
		un.sun_family = AF_UNIX;
		strcpy (un.sun_path,sockn);
		if (bind(fd,(struct sockaddr*)&un,sizeof(un))==-1){
			perror("bind");
		}else{
			chmod (sockn,0666);
			int code = listen (fd,10);
			if (code == -1){
				perror ("listen");
			}else{
				ret = fd;
			}	
		}
	}
	return ret;
}

static int aclfsd_sendfd(int sock, int fd)
{
	char bufctl[sizeof(struct cmsghdr)+1*sizeof(int)];
	struct cmsghdr *c = (struct cmsghdr*)bufctl;
	c->cmsg_len = sizeof(bufctl);
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_RIGHTS;
	int *fds = (int*)CMSG_DATA(c);
	fds[0] = fd;

	struct msghdr h;

	struct iovec iov[1];
	char buf[1000];
	strcpy (buf,"hellohello");
	iov[0].iov_base = buf;
	iov[0].iov_len = 10;

	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = (void*)bufctl;
	h.msg_controllen = sizeof(bufctl);
	h.msg_flags = 0;
	int ret = sendmsg (sock,&h,0);
	if (ret==-1){
		perror ("sendmsg");
	}
	return ret;
}

static bool do_update = false;

/*
	Must reread its config
*/
static void aclfsd_sighup(int)
{
	do_update = true;
}


struct ACLFSD_CONF {
	char *volume;
	char *abspath;
};

static int nbdef=0;
static ACLFSD_CONF *conf;

static void aclfsd_readconf (const char *configf)
{
	if (conf != NULL){
		for (int i=0; i<nbdef; i++){
			free (conf[i].volume);
			free (conf[i].abspath);
		}
		free (conf);
		nbdef = 0;
	}
	FILE *fin = fopen (configf,"r");
	if (fin != NULL){
		char buf[1000];
		conf = (ACLFSD_CONF*)malloc (100*sizeof(ACLFSD_CONF));
		while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
			if (buf[0] != '\0'){
				char *pt = buf;
				while (*pt > ' ') pt++;
				if (isspace(*pt)){
					*pt ++ = '\0';
					while (isspace(*pt)) pt++;
					char *end = pt + strlen(pt)-1;
					if (*end == '\n') *end = '\0';
					conf[nbdef].volume = strdup(buf);
					conf[nbdef].abspath = strdup(pt);
					nbdef++;
				}
			}
		}
		fclose (fin);
	}
}

static void aclfsd_sendfail (int fd)
{
	ACL_REPLY rep;
	rep.status = ACL_STATUS_FAIL;
	rep.errcode = errno;
	rep.len = 0;
	write (fd,&rep,sizeof(rep));
}

/*
	Send an access violation error
*/
static void aclfsd_sendperm (int fd)
{
	ACL_REPLY rep;
	rep.status = ACL_STATUS_FAIL;
	rep.errcode = EPERM;
	rep.len = 0;
	write (fd,&rep,sizeof(rep));
}

static void aclfsd_sendok (int fd)
{
	ACL_REPLY rep;
	rep.status = ACL_STATUS_OK;
	rep.errcode = 0;
	rep.len = 0;
	write (fd,&rep,sizeof(rep));
}

static void aclfsd_sendstat (int fd, struct stat &st)
{
	struct {
		ACL_REPLY rep;
		struct stat st;
	}s;
	s.rep.status = ACL_STATUS_STAT;
	s.rep.errcode = 0;
	s.rep.len = sizeof(struct stat);
	memcpy (&s.st,&st,sizeof(s.st));
	write (fd,&s,sizeof(s));
}

static void aclfsd_senddata (int fd, const void *p, int len)
{
	ACL_REPLY_DATA rep;
	rep.status = ACL_STATUS_DATA;
	rep.errcode = 0;
	rep.len = len;
	memcpy (&rep.data,p,len);
	write (fd,&rep,sizeof(ACL_REPLY)+len);
}

/*
	Convert the request to an absolute path by matching a volume
	definition. Send failure to the client if there is no matching
	volume or the request is invalid (.. in path)

	Return -1 if this is not a valid request.
*/
static int aclfsd_translate (
	int fd,
	char *path,
	char abspath[PATH_MAX],
	OPER_CONTEXT &oper,
	ACL_UCRED *uc)
{
	int ret = -1;
	if (path[0] == '/') path++;
	char *pt = strchr(path,'/');
	if (pt != NULL)	*pt++ = '\0';
	for (int i=0; i<nbdef; i++){
		if (strcmp(conf[i].volume,path)==0){
			if (pt == NULL){
				strcpy (abspath,conf[i].abspath);
				aclrights_get (uc->uid,uc->gid,uc->pid,abspath,".",oper);
				ret = 0;
			}else if (strstr(pt,"..")==NULL){
				snprintf (abspath,PATH_MAX-1,"%s/%s",conf[i].abspath,pt);
				pt = strrchr(abspath,'/');
				*pt = '\0';
				aclrights_get (uc->uid,uc->gid,uc->pid,abspath,pt+1,oper);
				*pt = '/';
				ret = 0;
			}
			break;
		}
	}
	if (ret == -1){
		aclfsd_sendfail (fd);
	}
	return ret;
}

/*
	Convert the request to an absolute path by matching a volume
	definition. Send failure to the client if there is no matching
	volume or the request is invalid (.. in path)

	Return -1 if this is not a valid request.
*/
static int aclfsd_translate (
	int fd,
	ACL_REQUEST *req,
	char abspath[PATH_MAX],
	OPER_CONTEXT &oper,
	ACL_UCRED *uc)
{
	return aclfsd_translate (fd,req->paths,abspath,oper,uc);
}

/*
	Return true if a file or directory exist
*/
bool file_exist (const char *fname)
{
	struct stat st;
	return stat(fname,&st)!=-1;
}

/*
	Read a request from one client
	Every request must contain a credential, to be added later
*/
static int aclfsd_doreq (int fd)
{
	char bufctl[sizeof(struct cmsghdr)+sizeof(ACL_UCRED)];
	memset (bufctl,-1,sizeof(bufctl));
	struct iovec iov[1];
	char buf[1000];
	iov[0].iov_base = buf;
	iov[0].iov_len = sizeof(buf);
	struct msghdr h;
	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = bufctl;
	h.msg_controllen = sizeof(bufctl);
	struct cmsghdr *c = (struct cmsghdr*) bufctl;
	c->cmsg_len = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_level = 0;
	c->cmsg_type = 0;

	n_debug ("msg len %d %d\n",c->cmsg_len,h.msg_controllen);


	h.msg_flags = 0;
	int len = recvmsg (fd,&h,0);
	int ret = -1;
	if (c->cmsg_level == SOL_SOCKET
		&& c->cmsg_type == SCM_CREDENTIALS){
		if (len>0){
			#if 0
				printf ("message recu flags %d len %d %d %u\n",h.msg_flags,len
					,h.msg_controllen,sizeof(bufctl));
				printf ("iov.len %d\n",iov[0].iov_len);
				printf ("msg_len %d msg_level %d msg_type %d\n",c->cmsg_len
					,c->cmsg_level
					,c->cmsg_type);
			#endif
			ACL_UCRED *uc = (ACL_UCRED*)CMSG_DATA(c);
			n_debug ("len %d msg %d %d %04x uc %d %d %d\n",len
				,c->cmsg_level,c->cmsg_type,h.msg_flags
				,uc->pid,uc->uid,uc->gid);
			ACL_REQUEST *req = (ACL_REQUEST*)buf;
			buf[len] = '\0';
			switch (req->cmd){
			case ACL_OPEN:
				{
				n_debug ("open :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					int mode = req->mode;
					n_debug ("open abspath :%s: rights %08x mode %04x\n"
						,abspath,oper,mode);
					bool ok = true;
					struct stat st;
					bool file_exist = stat(abspath,&st)!=-1;
					if (!file_exist
						&& (mode & O_CREAT)!=0
						&& !chk_r(oper.rights,may_create)) ok = false;
					if (file_exist
						&& (mode & O_TRUNC)!=0
						&& !chk_r(oper.rights,may_trunc)) ok = false;
					if ((mode & O_APPEND)!=0 && !chk_r(oper.rights,may_append)) ok = false;
					if ((mode & O_WRONLY)!=0){
						if (!chk_r(oper.rights,may_write)) ok = false;
					}else if ((mode & O_RDWR)!=0){
						if (!chk_r(oper.rights,may_write)
							|| !chk_r(oper.rights,may_read)) ok = false;
					}else if (!chk_r(oper.rights,may_read)){
						ok = false;
					}

					if (file_exist
						&& ((mode & O_WRONLY) != 0 || (mode & O_RDWR) != 0)
						&& !chk_r(oper.rights,may_overwrite)) ok = false;
					if (!ok){
						aclfsd_sendperm(fd);
					}else{
						int handle = open (abspath,req->mode,req->perm);
						if (handle != -1){
							if (!file_exist){
								fchown (handle,oper.files.uid,oper.files.gid);
								if (oper.files.mode != -1){
									fchmod (handle,oper.files.mode);
								}
							}
							aclfsd_sendfd (fd,handle);
							close (handle);
						}else{
							aclfsd_sendfail(fd);
						}
					}
				}
				}
				break;
			case ACL_STAT:
				{
				n_debug ("stat :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("stat abspath :%s:\n",abspath);
					struct stat st;
					if (!chk_r(oper.rights,may_stat)){
						aclfsd_sendperm(fd);
					}else if (stat(abspath,&st)==-1){
						aclfsd_sendfail(fd);
					}else{
						aclfsd_sendstat (fd,st);
					}
				}
				}
				break;
			case ACL_LSTAT:
				{
				n_debug ("lstat :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("lstat abspath :%s:\n",abspath);
					struct stat st;
					if (!chk_r(oper.rights,may_stat)){
						aclfsd_sendperm(fd);
					}else if (lstat(abspath,&st)==-1){
						aclfsd_sendfail(fd);
					}else{
						aclfsd_sendstat (fd,st);
					}
				}
				}
				break;
			case ACL_UNLINK:
				{
				n_debug ("unlink :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("unlink abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_unlink)){
						aclfsd_sendperm(fd);
					}else if (unlink(abspath)!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
				}
				break;
			case ACL_UTIMES:
				{
				n_debug ("utimes :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("utimes abspath :%s: %d %d\n",abspath,req->mode,req->perm);
					if (!chk_r(oper.rights,may_utimes)){
						aclfsd_sendperm(fd);
					}else{
						int ok = -1;
						if (req->mode == -1 && req->perm == -1){
							ok = utimes (abspath,NULL);
						}else{
							struct timeval tvp[2];
							tvp[0].tv_sec = req->mode;
							tvp[1].tv_sec = req->perm;
							tvp[0].tv_usec = 0;
							tvp[1].tv_usec = 0;
							ok = utimes (abspath,tvp);
						}
						if (ok!=-1){
							aclfsd_sendok (fd);
						}else{
							aclfsd_sendfail (fd);
						}
					}
				}
				}
				break;
			case ACL_ACCESS:
			case ACL_EUIDACCESS:	// For now, setuid program do not
									// work with virtualfs.
				{
				n_debug ("access :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					bool ok = true;
					int type = req->mode;
					if ((type & F_OK) != 0 && !chk_r(oper.rights,may_stat)) ok = false;
					if ((type & R_OK) != 0 && !chk_r(oper.rights,may_read)) ok = false;
					if ((type & W_OK) != 0 && !chk_r(oper.rights,may_write)) ok = false;
					if ((type & X_OK) != 0) ok = false;
					n_debug ("access abspath :%s: %d ok=%d\n",abspath,req->mode,ok);
					if (ok){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendperm (fd);
					}
				}
				}
				break;
			case ACL_RMDIR:
				{
				n_debug ("rmdir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("rmdir abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_rmdir)){
						aclfsd_sendperm(fd);
					}else if (rmdir(abspath)!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
				}
				break;
			case ACL_MKDIR:
				{
				n_debug ("mkdir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("mkdir abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_mkdir)){
						aclfsd_sendperm(fd);
					}else if (mkdir(abspath,req->mode)!=-1){
						chown (abspath,oper.dirs.uid,oper.dirs.gid);
						if (oper.dirs.mode != -1){
							chmod (abspath,oper.dirs.mode);
						}
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
				}
				break;
			case ACL_CHDIR:
				{
				// chdir means, may chdir
				n_debug ("chdir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("chdir abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_readdir)){
						aclfsd_sendperm(fd);
					}else{
						struct stat st;
						if (stat(abspath,&st)!=-1 && S_ISDIR(st.st_mode)){
							aclfsd_sendok (fd);
						}else{
							aclfsd_sendfail (fd);
						}
					}
				}
				}
				break;
			case ACL_OPENDIR:
				{
				n_debug ("opendir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("opendir abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_readdir)){
						aclfsd_sendperm(fd);
					}else{
						DIR *dir = opendir (abspath);
						if (dir != NULL){
							int tbfd[2];
							if (pipe(tbfd)!=-1){
								aclfsd_sendfd (fd,tbfd[0]);
								if (fork()==0){
									struct dirent *ent;
									while ((ent=readdir(dir))!=NULL){
										write (tbfd[1],ent,sizeof(struct dirent));
									}
									closedir (dir);
									_exit (0);
								}
								close (tbfd[0]);
								close (tbfd[1]);
							}
							closedir(dir);
						}else{
							aclfsd_sendfail (fd);
						}
					}
				}
				}
				break;
			case ACL_RENAME:
			case ACL_LINK:
			case ACL_SYMLINK:
				{
					char *newp = req->paths;
					while (*newp != '\0'
						&& (int)(newp-req->paths) < PATH_MAX) newp++;
					newp++;
					n_debug ("rename/link/symlink :%s: :%s:\n",req->paths,newp);
					char old_abspath[PATH_MAX];
					OPER_CONTEXT oper;
					if (req->cmd == ACL_SYMLINK
						|| aclfsd_translate (fd,req,old_abspath,oper,uc)!=-1){
						char new_abspath[PATH_MAX];
						OPER_CONTEXT new_oper;
						if (aclfsd_translate (fd,newp,new_abspath
							,new_oper,uc)!=-1){
							if (req->cmd == ACL_RENAME){
								n_debug ("rename abspath :%s: :%s:\n",old_abspath,new_abspath);
								if (!chk_r(oper.rights,may_rename)){
									aclfsd_sendperm(fd);
								}else if(!chk_r(new_oper.rights,may_unlink)
									&& file_exist (new_abspath)){
									aclfsd_sendperm(fd);
								}else if (rename(old_abspath,new_abspath)!=-1){
									aclfsd_sendok (fd);
								}else{
									aclfsd_sendfail (fd);
								}
							}else if (req->cmd == ACL_LINK){
								n_debug ("link abspath :%s: :%s:\n",old_abspath,new_abspath);
								if (!chk_r(oper.rights,may_link)
									|| !chk_r(oper.rights,may_link)){
									aclfsd_sendperm(fd);
								}else if (link(old_abspath,new_abspath)!=-1){
									aclfsd_sendok (fd);
								}else{
									aclfsd_sendfail (fd);
								}
							}else{
								n_debug ("symlink path :%s: :%s:\n"
									,req->paths,new_abspath);
								if (!chk_r(new_oper.rights,may_symlink)){
									aclfsd_sendperm(fd);
								}else if (symlink(req->paths,new_abspath)!=-1){
									aclfsd_sendok (fd);
								}else{
									aclfsd_sendfail (fd);
								}
							}
						}
					}
				}
				break;
			case ACL_CHMOD:
				{
				n_debug ("chmod :%s: %o\n",req->paths,req->mode);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("chmod abspath :%s:\n",abspath);
					struct stat st;
					if (stat(abspath,&st)==-1){
						if (!chk_r(oper.rights,may_stat)){
							// The file does not exist and the user is not
							// supposed to know
							aclfsd_sendperm (fd);
						}else{
							aclfsd_sendfail (fd);
						}
					}else if ((req->mode & 07000) != 0
						&& !chk_r(oper.rights,may_chmodsuid)){
						aclfsd_sendperm(fd);
					}else if (st.st_uid != (uid_t)uc->uid
						&& !chk_r(oper.rights,may_chmodany)){
						aclfsd_sendperm(fd);
					}else{
						if (!chk_r(oper.rights,may_chmod)
							&& !chk_r(oper.rights,may_chmodany)){
							aclfsd_sendperm(fd);
						}else if (chmod(abspath,req->mode)!=-1){
							aclfsd_sendok (fd);
						}else{
							aclfsd_sendfail (fd);
						}
					}
				}
				}
				break;
			case ACL_CHOWN:
				{
				n_debug ("chown :%s:\n",req->paths);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					bool chown_ok  = chk_r(oper.rights,may_chown);
					bool chown_any = chk_r(oper.rights,may_chownany);
					int newuid = req->mode;
					int newgid = req->perm;
					if (newuid != -1  || newgid != -1){
						if (!chown_any){
							aclfsd_sendperm(fd);
						}else if (chown(abspath,newuid,newgid)!=-1){
							aclfsd_sendok(fd);
						}else{
							aclfsd_sendfail (fd);
						}
					}else if (!chown_ok){
						aclfsd_sendperm(fd);
					}else{
						if (chown(abspath,newuid,newgid)!=-1){
							aclfsd_sendok(fd);
						}else{
							aclfsd_sendfail (fd);
						}
					}
				}
				}
				break;
			case ACL_MKNOD:
				{
				n_debug ("mknod :%s: %o %o\n",req->paths,req->mode,req->perm);
				char abspath[PATH_MAX];
				OPER_CONTEXT oper;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("chmod abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_mknod)){
						aclfsd_sendperm(fd);
					}else if (mknod (abspath,req->mode,req->perm)!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
				}
				break;
			case ACL_READLINK:
				{
				n_debug ("readlink :%s:\n",req->paths);
				char abspath[PATH_MAX],linkpath[PATH_MAX];
				OPER_CONTEXT oper;
				int len;
				if (aclfsd_translate (fd,req,abspath,oper,uc)!=-1){
					n_debug ("readlink abspath :%s:\n",abspath);
					if (!chk_r(oper.rights,may_readlink)){
						aclfsd_sendperm(fd);
					}else if ((len=readlink(abspath,linkpath
						,sizeof(linkpath)-1))!=-1){
						aclfsd_senddata (fd,linkpath,len);
					}else{
						aclfsd_sendfail (fd);
					}
				}
				}
				break;
			case ACL_RELOADACL:
				{
				n_debug ("reloadacl :%s: uid=%d\n",req->paths,uc->uid);
				if (uc->uid == 0){
					aclrights_reload (req->paths);
					aclfsd_sendok (fd);
				}else{
					aclfsd_sendperm(fd);
				}
				}
				break;
			case ACL_RELOADCONF:
				{
				n_debug ("reloadconf\n");
				if (uc->uid == 0){
					do_update = true;
					aclfsd_sendok (fd);
				}else{
					aclfsd_sendperm(fd);
				}
				}
				break;
			default:
				aclfsd_sendfail(fd);
			}
			ret = 0;		
		}
	}
	return ret;
}
		



/*
	Process all request and connections. Infinite loop
*/
static void aclfsd_loop (int sock, const char *configf)
{
	int tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = sock+1;
	FD_SET (sock,&fdin);
	signal (SIGHUP,aclfsd_sighup);
	signal (SIGCHLD,SIG_IGN);
	signal (SIGPIPE,SIG_IGN);
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		if (select (maxfd,&in,NULL,NULL,NULL)!=-1){
			int orig_nbfd = nbfd;
			if (FD_ISSET (sock,&in)){
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_UNIX;
				int s = accept (sock,(struct sockaddr*)&unc,&len);
				int val = 1;
				if (setsockopt (s,SOL_SOCKET,SO_PASSCRED,&val,sizeof(val))!=0){
					perror ("setsockopt");
				}

				tbfd[nbfd++] = s;
				FD_SET (s,&fdin);
				if (s >= maxfd) maxfd = s + 1;
			}
			int *ptfd = tbfd;
			bool redo_fdin = false;
			for (int i=0; i< orig_nbfd; i++, ptfd++){
				int fd = *ptfd;
				if (FD_ISSET(fd,&in)){
					if (aclfsd_doreq (fd) == -1){
						close (fd);
						redo_fdin = true;
						*ptfd = -1;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = sock+1;
				FD_SET (sock,&fdin);
				ptfd = tbfd;
				int *copyfd = tbfd;
				for (int i=0; i<nbfd; i++, ptfd++){
					int fd = *ptfd;
					if (fd != -1){
						FD_SET (fd,&fdin);
						if (fd >= maxfd) maxfd = fd + 1;
						*copyfd++ = fd;
					}
				}
				nbfd = (int)(copyfd - tbfd);
			}
		}
		if (do_update){
			do_update = false;
			aclfsd_readconf (configf);
		}
	}
}
extern "C" char *stpcpy (char *, const char *);

static void aclfsd_reload (const char *dir)
{
	if (aclrights_checkacl(dir)!=-1){
		ACL_REQUEST req;
		char *end = stpcpy (req.paths,dir) + 1;
		int len_request = (int)(end-req.paths);
		req.mode = 0;
		req.perm = 0;
		req.cmd = ACL_RELOADACL;
		acllib_connect();
		acllib_sendreq (req,len_request);
		acllib_receive();
	}
}

static void aclfsd_reloadconf ()
{
	ACL_REQUEST req;
	req.mode = 0;
	req.perm = 0;
	req.cmd = ACL_RELOADCONF;
	n_debug ("Connecting\n");
	acllib_connect();
	n_debug ("is connected\n");
	acllib_sendreq (req,0);
	n_debug ("message sent\n");
	acllib_receive();
	n_debug ("message received\n");

}

int main (int argc, char *argv[])
{
	int ret = -1;
	const char *sockpath = ACLFS_SOCKET;
	const char *pidfile = "/var/run/aclfsd.pid";
	const char *configf = "/etc/aclfsd.conf";
	bool testdone = false;
	bool fore = false;		// Stay in foreground
	int i;
	for (i=1; i<argc; i++){
		if (strcmp(argv[i],"--socket")==0){
			i++;
			sockpath = argv[i];
		}else if (strcmp(argv[i],"--pidfile")==0){
			i++;
			pidfile = argv[i];
		}else if (strcmp(argv[i],"--config")==0){
			i++;
			configf = argv[i];
		}else if (strcmp(argv[i],"--reload")==0){
			i++;
			aclfsd_reload (argv[i]);
			testdone = true;
		}else if (strcmp(argv[i],"--reloadconf")==0){
			aclfsd_reloadconf ();
			testdone = true;
		}else if (strcmp(argv[i],"--testacl")==0){
			i++;
			if (i != argc-4){
				usage();
			}else{
				aclrights_testacl (argv[i],argv[i+1],argv[i+2],argv[i+3]);
				i+=3;
			}
			testdone = true;
		}else if (strcmp(argv[i],"--debug")==0){
			debug = true;
		}else if (strcmp(argv[i],"--nodaemon")==0){
			fore = true;
		}else if (strcmp(argv[i],"--help")==0){
			usage();
			exit (-1);
		}else{
			usage();
			exit (-1);
		}
	}


	if (i == argc && !testdone){
		if (fore || fork()==0){
			openlog ("aclfsd",LOG_CONS,LOG_AUTHPRIV);
			{
				FILE *fout = fopen (pidfile,"w");
				if (fout != NULL){
					fprintf (fout,"%d\n",getpid());
					fclose (fout);
				}
			}
			aclfsd_readconf(configf);
			int fd = aclfsd_server(sockpath);
			if (fd != -1){
				chdir ("/");
				aclfsd_loop(fd,configf);
				ret = 0;
			}
		}else{
			ret = 0;
		}
	}
	return ret;
}


