#ifndef ACLFSD_H
#define ACLFSD_H

#include <stdio.h>
#include <sys/types.h>
#include <regex.h>

enum ACL_RIGHTS {
	acl_rights_none=0,
	acl_rights_all=0xffffffff,
	may_overwrite=1<<0,
	may_write=1<<1,
	may_append=1<<2,
	may_trunc=1<<3,
	may_read=1<<4,
	may_create=1<<5,
	may_mkdir=1<<6,
	may_unlink=1<<7,
	may_rmdir=1<<8,
	may_stat=1<<9,
	may_readdir=1<<10,
	may_rename=1<<11,
	may_utimes=1<<12,
	may_link=1<<13,
	may_symlink=1<<14,
	may_chmod=1<<15,
	may_readlink=1<<16,
	may_chown=1<<17,
	may_chownany=1<<18,
	may_chmodany=1<<19,
	may_chmodsuid=1<<20,
	may_mknod=1<<21,
	may_exec=1<<22,
};

struct CREATE_MODE {
	bool is_filled;	// This was filled with effective values
	uid_t uid;		// User ID used to create files or directories
						// -1 means to pick the identity of the caller
	gid_t gid;		// Group ID
	int   mode;		// Creation mode
};

struct OPER_CONTEXT{
	int rights;
	CREATE_MODE files;
	CREATE_MODE dirs;
};

enum ACLTOK_TYPE {
	ACLTOK_EOF,
	ACLTOK_OPNBRAC, ACLTOK_CLSBRAC, ACLTOK_COLON, ACLTOK_SCOLON,
	ACLTOK_COMMA,
	ACLTOK_GROUP, ACLTOK_LIST, ACLTOK_ID, ACLTOK_NOT, ACLTOK_DASH
};

class ACLTOK{
	FILE *fin;
	char *fname;
	char line[1000];
	char *ptline;
	int noline;
	bool oneerror;
	bool outlog;
public:
	char token[1000];
	ACLTOK_TYPE type;
	/*~PROTOBEG~ ACLTOK */
public:
	ACLTOK (const char *_fname, FILE *_fin, bool _outlog);
	int cmp (const char *s);
	int getline (void)const;
	void logerror (const char *ctl, ...);
	void logsyntax (void);
	void logwarning (const char *ctl, ...);
	int next (ACLTOK_TYPE expect);
	ACLTOK_TYPE next (void);
	void seterror (void);
	bool waserror (void);
	~ACLTOK (void);
	/*~PROTOEND~ ACLTOK */
};


enum ITEM_TYPE { ITEM_UID, ITEM_GID, ITEM_LIST, ITEM_ALL};

class ACL_LIST;
class ACL_LISTS;

class ACL_ITEM{
public:
	int id;
	ACL_ITEM *next;
	ITEM_TYPE type;
	const ACL_LIST *list;
	/*~PROTOBEG~ ACL_ITEM */
public:
	ACL_ITEM (ACLTOK&tok, const ACL_LISTS&lists);
	/*~PROTOEND~ ACL_ITEM */
};

class ACL_LIST{
public:
	char *name;
	ACL_LIST *next;
	ACL_ITEM *first;
	/*~PROTOBEG~ ACL_LIST */
public:
	ACL_LIST (ACLTOK&tok, ACL_LISTS&lists);
	ACL_LIST (void);
	void add (ACL_ITEM *item);
	void dump (const char *indent, FILE *fout);
	bool is_member (int uid, int gid)const;
	void parse (ACLTOK&tok, ACL_LISTS&lists);
	~ACL_LIST (void);
	/*~PROTOEND~ ACL_LIST */
};

class ACL_LISTS{
public:
	ACL_LIST *first;
	/*~PROTOBEG~ ACL_LISTS */
public:
	ACL_LISTS (void);
	void add (ACL_LIST *list);
	~ACL_LISTS (void);
	/*~PROTOEND~ ACL_LISTS */
};

class ACL_ITEMSTR{
public:
	ACL_ITEMSTR *next;
	char *str;
	/*~PROTOBEG~ ACL_ITEMSTR */
public:
	ACL_ITEMSTR (const char *_str, ACL_ITEMSTR *_next);
	~ACL_ITEMSTR (void);
	/*~PROTOEND~ ACL_ITEMSTR */
};

class ACL_LISTSTR{
public:
	ACL_ITEMSTR *first;
	/*~PROTOBEG~ ACL_LISTSTR */
public:
	ACL_LISTSTR (void);
	void add (const char *_str);
	~ACL_LISTSTR (void);
	/*~PROTOEND~ ACL_LISTSTR */
};

class ACL_CONTEXT {
public:
	char *name;
	ACL_CONTEXT *next;
	ACL_LIST grant;			// Those who received those rights
	ACL_LIST deny;			// Those who should not
	bool oper_set;			// There is valid data in oper
	OPER_CONTEXT oper;
	ACL_LISTSTR programs;
	ACL_LISTSTR modules;
	/*~PROTOBEG~ ACL_CONTEXT */
public:
	ACL_CONTEXT (ACLTOK&tok, ACL_LISTS&lists);
	/*~PROTOEND~ ACL_CONTEXT */
};

class ACL_CONTEXTS {
public:
	ACL_CONTEXT *first;
	/*~PROTOBEG~ ACL_CONTEXTS */
public:
	ACL_CONTEXTS (void);
	void add (ACL_CONTEXT *ctx);
	~ACL_CONTEXTS (void);
	/*~PROTOEND~ ACL_CONTEXTS */
};


class ACL_DIRITEM{
public:
	bool is_regex;
	regex_t reg;
	char *pattern;
	ACL_DIRITEM *next;
	ACL_CONTEXTS contexts;
	OPER_CONTEXT oper_default;
	/*~PROTOBEG~ ACL_DIRITEM */
public:
	ACL_DIRITEM (ACLTOK&tok, ACL_LISTS&lists);
	~ACL_DIRITEM (void);
	/*~PROTOEND~ ACL_DIRITEM */
};

class ACL_DIRITEMS{
public:
	ACL_DIRITEM *first;
	/*~PROTOBEG~ ACL_DIRITEMS */
public:
	ACL_DIRITEMS (void);
	void add (ACL_DIRITEM *d);
	~ACL_DIRITEMS (void);
	/*~PROTOEND~ ACL_DIRITEMS */
};

class ACL_DIR{
public:
	char *path;
	ACL_DIR *next;
	ACL_DIR *inherit;
	ACL_LISTS lists;
	ACL_DIRITEMS items;
	bool no__acl;		// Is there an ..acl file in this directory
						// If not, it inherits the one for the parent
	bool aclsyntaxok;	// Is the ..acl file valid
	OPER_CONTEXT oper_default;
	/*~PROTOBEG~ ACL_DIR */
public:
	ACL_DIR (const char *_path,
		 ACL_DIR *_next,
		 bool outlog);
	void dump (FILE *fout);
	void getrights (int uid,
		 int gid,
		 int pid,
		 const char *file,
		 OPER_CONTEXT&oper);
	/*~PROTOEND~ ACL_DIR */
};

typedef void *(*initctx_f)(const char *args[]);
typedef	void (*endctx_f)(void *);
typedef	int  (*testctx_f)(
	void *, OPER_CONTEXT &, uid_t uid, gid_t gid, pid_t pid);


class ACLMODULE{
	void *handle;
public:
	char *name;
	initctx_f initctx;
	endctx_f endctx;
	testctx_f testctx;
	/*~PROTOBEG~ ACLMODULE */
public:
	ACLMODULE (const char *_name);
	bool is_ok (void)const;
	~ACLMODULE (void);
	/*~PROTOEND~ ACLMODULE */
};

inline bool chk_r (int rights, ACL_RIGHTS one)
{
	return (rights & one) != 0;
}


#include "aclfsd.p"

#endif

