#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include "aclfsd.h"

PUBLIC ACLTOK::ACLTOK(
	const char *_fname,
	FILE *_fin,
	bool _outlog)	// Error message in syslog
{
	outlog = _outlog;
	fname = strdup(_fname);		// Used to signal error
	fin = _fin;
	line[0] = '\0';
	token[0] = '\0';
	noline = 0;
	ptline = line;
	oneerror = false;
}

PUBLIC ACLTOK::~ACLTOK()
{
	free (fname);
}

PUBLIC int ACLTOK::getline() const
{
	return noline;
}

/*
	Put the next token in token[] and return its type.
	Return ACLTOK_EOF if enf of file.
*/
PUBLIC ACLTOK_TYPE ACLTOK::next ()
{
	type = ACLTOK_EOF;
	token[0] = '\0';
	while (!oneerror){
		while (*ptline <= ' ' && *ptline != '\0') ptline++;
		if (*ptline == '\0' || *ptline == '#'){
			if (fgets(line,sizeof(line)-1,fin)==NULL){
				break;
			}
			noline++;
			ptline = line;
		}else if (*ptline == '{'){
			type = ACLTOK_OPNBRAC;
			ptline++;
			break;
		}else if (*ptline == '}'){
			type = ACLTOK_CLSBRAC;
			ptline++;
			break;
		}else  if(*ptline == ','){
			type = ACLTOK_COMMA;
			ptline++;
			break;
		}else if (*ptline == ';'){
			type = ACLTOK_SCOLON;
			ptline++;
			break;
		}else if (*ptline == '!'){
			type = ACLTOK_NOT;
			ptline++;
			break;
		}else{
			if (*ptline == '@'){
				type = ACLTOK_GROUP;
				ptline++;
			}else if (*ptline == '&'){
				type = ACLTOK_LIST;
				ptline++;
			}else{
				type = ACLTOK_ID;
			}
			char *pttok = token;
			while (*ptline > ' '){
				if (*ptline == '{'
					|| *ptline == '}'
					|| *ptline == ','
					|| *ptline == '!'
					|| *ptline == ';'){
					break;
				}
				*pttok++ = *ptline++;
			}
			*pttok = '\0';
			if (type == ACLTOK_ID && token[0] == '-' && token[1] == '\0'){
				type = ACLTOK_DASH;
			}
			break;
		}
	}
	return type;
}

// Pick a token and expect a specific type
PUBLIC int ACLTOK::next (ACLTOK_TYPE expect)
{
	ACLTOK_TYPE t = next();
	int ret = 0;
	if (t != expect){
		logerror ("Invalid token\n");
		ret = -1;
	}
	return ret;
}

// Record that there was an error during parsing
PUBLIC void ACLTOK::seterror()
{
	oneerror = true;
}

// Return true if there were one error.
PUBLIC bool ACLTOK::waserror()
{
	return oneerror;
}

/*
	Signal an important error, but not enough to prevent usage of the ..acl
	file
*/
PUBLIC void ACLTOK::logwarning(
	const char *ctl,
	...)
{
	va_list list;
	va_start (list,ctl);
	if (outlog){
		char buf[1000];
		vsnprintf (buf,sizeof(buf)-1,ctl,list);
		syslog (LOG_ERR,"file %s, line %d: %s",fname,noline,buf);
	}else{
		fprintf (stderr,"File %s, line %d: ",fname,noline);
		vfprintf (stderr,ctl,list);
		fputc ('\n',stderr);
	}
	va_end (list);
}

/*
	Signal an important fatal error. The ..acl won't be useable and
	all access to the directory will be denied.
*/
PUBLIC void ACLTOK::logerror(
	const char *ctl,
	...)
{
	va_list list;
	va_start (list,ctl);
	if (outlog){
		char buf[1000];
		vsnprintf (buf,sizeof(buf)-1,ctl,list);
		syslog (LOG_ERR,"file %s, line %d: %s",fname,noline,buf);
	}else{
		fprintf (stderr,"File %s, line %d: ",fname,noline);
		vfprintf (stderr,ctl,list);
		fputc ('\n',stderr);
	}
	va_end (list);
	seterror();
}

/*
	Log a syntax error without much detail
*/
PUBLIC void ACLTOK::logsyntax()
{
	logerror("Syntax error");
}

/*
	Compare a string with the last extracted token.
	Return the output of strcmp().
*/
PUBLIC int ACLTOK::cmp(const char *s)
{
	return strcmp(token,s);
}

