#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include "cmdsock.h"
#include "remaudio.h"
#include <linux/soundcard.h>
#include <linux/cdrom.h>

static bool debug=false;

static void usage()
{
	fprintf (stderr,"remaudioserv [ --debug ] [ --silentmount ]\n");
	exit (-1);
}

void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vprintf (ctl,list);
		va_end (list);
	}
}

/*
	Read exactly len byte from the socket, using multiple read if needed
	Return the number of byte read or -1 if error
*/
static int audioserv_bufread (int cli, void *buf, int len)
{
	char *pbuf = (char*)buf;
	int sofar = 0;
	while (sofar < len){
		int nb=read(cli,pbuf+sofar,len-sofar);
		if (nb > 0){
			sofar += nb;
		}else if (errno != EAGAIN){
			break;
		}
	}
	return sofar != len ? -1 : len;
}

static void audioserv_output (int cli, const char *fname, int mode)
{
	if (fork()==0){
		int fd = open (fname,mode,0);
		n_debug ("opening %s -> %d : %d %d <> %d %d\n"
			,fname,fd,mode,errno,O_WRONLY,EBUSY);
		if (fd != -1 || (mode == O_WRONLY && errno == EBUSY)){
			REMAUDIO_CTL ctl;
			ctl.len = 0;
			ctl.p1 = htonl(fd);
			write (cli,&ctl,sizeof(ctl));
			while (audioserv_bufread(cli,&ctl,sizeof(ctl))==sizeof(ctl)){
				ctl.cmd = ntohs (ctl.cmd);
				ctl.request = ntohs (ctl.request);
				ctl.len = ntohl  (ctl.len);
				ctl.p1  = ntohl  (ctl.p1);
				ctl.p2  = ntohl  (ctl.p2);
				ctl.p3  = ntohl  (ctl.p3);
				if (ctl.cmd == NETCTL_WRITE){
					char buf[ctl.len];
					audioserv_bufread (cli,buf,ctl.len);
					if (fd == -1){
						n_debug ("Attempting to open the busy %s\n"
							,fname);
						fd = open (fname,mode,0);
						if (fd != -1 && errno != EBUSY) break;
					}
					if (fd != -1){
						int written = write (fd,buf,ctl.len);
						n_debug ("writing %d bytes -> %s\n",ctl.len,written);
						ctl.len = written;
					}else{
						syslog (LOG_ERR,"Can't write to %s %d bytes: busy\n"
							,fname,ctl.len);
					}
					ctl.p1 = 0;
					write (cli,&ctl,sizeof(ctl));
				}else if (ctl.cmd == NETCTL_READ){
					char buf[ctl.len];
					int len = read (fd,buf,ctl.len);
					n_debug ("read %s len = %d %d\n",fname,ctl.len,len);
					ctl.len = len;
					write (cli,&ctl,sizeof(ctl));
					if (ctl.len > 0) write (cli,buf,ctl.len);
				}else if (ctl.cmd == NETCTL_LSEEK){
					ctl.len = lseek (fd,ctl.p1,ctl.p2);
					write (cli,&ctl,sizeof(ctl));
				}else if (ctl.cmd == NETCTL_IOCTL){
					int request = ctl.request;
					struct cdrom_subchnl sub;
					struct cdrom_tochdr tochdr;
					struct cdrom_tocentry tocentry;
					struct cdrom_msf msf;
					struct cdrom_volctrl volctrl;
					n_debug ("ioctl %s %08x %08x\n",fname,ctl.request
						,SNDCTL_DSP_SAMPLESIZE);
					if (ctl.request == SNDCTL_DSP_SAMPLESIZE){
						n_debug ("Sample size %d\n",ctl.p1);
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SNDCTL_DSP_SPEED){
						n_debug ("Speed %d\n",ctl.p1);
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SNDCTL_DSP_STEREO){
						n_debug ("Stereo %d\n",ctl.p1);
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SNDCTL_DSP_RESET){
						n_debug ("Reset\n");
						ctl.len = ioctl (fd,ctl.request,NULL,NULL,NULL);
					}else if (ctl.request == SNDCTL_DSP_GETCAPS){
						n_debug ("getcaps\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SOUND_MIXER_READ_CAPS){
						n_debug ("mixer read caps\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SOUND_MIXER_READ_DEVMASK){
						n_debug("mixer read caps\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SOUND_MIXER_READ_RECMASK){
						n_debug ("mixer read caps\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SOUND_MIXER_READ_STEREODEVS){
						n_debug ("mixer read caps\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SOUND_MIXER_READ_VOLUME
						|| ctl.request == SOUND_MIXER_READ_BASS
						|| ctl.request == SOUND_MIXER_READ_TREBLE
						|| ctl.request == SOUND_MIXER_READ_SYNTH
						|| ctl.request == SOUND_MIXER_READ_PCM
						|| ctl.request == SOUND_MIXER_READ_SPEAKER
						|| ctl.request == SOUND_MIXER_READ_LINE
						|| ctl.request == SOUND_MIXER_READ_MIC
						|| ctl.request == SOUND_MIXER_READ_CD
						|| ctl.request == SOUND_MIXER_READ_IMIX
						|| ctl.request == SOUND_MIXER_READ_ALTPCM
						|| ctl.request == SOUND_MIXER_READ_RECLEV
						|| ctl.request == SOUND_MIXER_READ_IGAIN
						|| ctl.request == SOUND_MIXER_READ_OGAIN
						|| ctl.request == SOUND_MIXER_READ_LINE1
						|| ctl.request == SOUND_MIXER_READ_LINE2
						|| ctl.request == SOUND_MIXER_READ_LINE3){
						n_debug ("read mixer value\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == SOUND_MIXER_WRITE_VOLUME
						|| ctl.request == SOUND_MIXER_WRITE_BASS
						|| ctl.request == SOUND_MIXER_WRITE_TREBLE
						|| ctl.request == SOUND_MIXER_WRITE_SYNTH
						|| ctl.request == SOUND_MIXER_WRITE_PCM
						|| ctl.request == SOUND_MIXER_WRITE_SPEAKER
						|| ctl.request == SOUND_MIXER_WRITE_LINE
						|| ctl.request == SOUND_MIXER_WRITE_MIC
						|| ctl.request == SOUND_MIXER_WRITE_CD
						|| ctl.request == SOUND_MIXER_WRITE_IMIX
						|| ctl.request == SOUND_MIXER_WRITE_ALTPCM
						|| ctl.request == SOUND_MIXER_WRITE_RECLEV
						|| ctl.request == SOUND_MIXER_WRITE_IGAIN
						|| ctl.request == SOUND_MIXER_WRITE_OGAIN
						|| ctl.request == SOUND_MIXER_WRITE_LINE1
						|| ctl.request == SOUND_MIXER_WRITE_LINE2
						|| ctl.request == SOUND_MIXER_WRITE_LINE3){
						n_debug ("set mixer value\n");
						ctl.len = ioctl (fd,ctl.request,&ctl.p1,NULL,NULL);
					}else if (ctl.request == CDROMSUBCHNL){
						audioserv_bufread(cli,&sub,sizeof(sub));
						ctl.len = ioctl (fd,ctl.request,&sub,NULL,NULL);
					}else if (ctl.request == CDROMREADTOCHDR){
						audioserv_bufread(cli,&tochdr,sizeof(tochdr));
						ctl.len = ioctl (fd,ctl.request,&tochdr,NULL,NULL);
					}else if (ctl.request == CDROMREADTOCENTRY){
						audioserv_bufread(cli,&tocentry,sizeof(tocentry));
						ctl.len = ioctl (fd,ctl.request,&tocentry,NULL,NULL);
					}else if (ctl.request == CDROMPLAYMSF){
						audioserv_bufread(cli,&msf,sizeof(msf));
						ctl.len = ioctl (fd,ctl.request,&msf,NULL,NULL);
					}else if (ctl.request == CDROMVOLCTRL
						|| ctl.request == CDROMVOLREAD){
						audioserv_bufread(cli,&volctrl,sizeof(volctrl));
						ctl.len = ioctl (fd,ctl.request,&volctrl,NULL,NULL);
					}else if (request == CDROMPAUSE
						|| request == CDROMRESUME
						|| request == CDROMSTOP
						|| request == CDROMSTART
						|| request == CDROMEJECT
						|| request == CDROMCLOSETRAY
						|| request == CDROMRESET){
						ctl.len = ioctl (fd,ctl.request,NULL,NULL,NULL);
					}

					ctl.request = errno;
					write (cli,&ctl,sizeof(ctl));
					if (request == CDROMSUBCHNL){
						write (cli,&sub,sizeof(sub));
					}else if (request == CDROMREADTOCHDR){
						write (cli,&tochdr,sizeof(tochdr));
					}else if (request == CDROMREADTOCENTRY){
						write (cli,&tocentry,sizeof(tocentry));
					}else if (request == CDROMPLAYMSF){
						write (cli,&msf,sizeof(msf));
					}else if (request == CDROMVOLCTRL
						|| request == CDROMVOLREAD){
						write (cli,&volctrl,sizeof(volctrl));
					}
				}
			}
		}else{
			syslog (LOG_ERR,"Can't open %s (%s)",fname,strerror(errno));
			write (cli,"notok",6);
		}
		n_debug ("audioserv (%s) %s: quitting\n"
			,mode==O_WRONLY ? "output" : "input"
			,fname);
		_exit (0);
	}
}
static bool audioserv_ismounted(const char *volume)
{
	bool ret = false;
	char mpoint[PATH_MAX];
	snprintf (mpoint,sizeof(mpoint)-1,"/mnt/%s",volume);
	FILE *fin = fopen ("/proc/mounts","r");
	if (fin != NULL){
		char buf[2*PATH_MAX];
		while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
			char dev[PATH_MAX],dir[PATH_MAX],stuff[PATH_MAX];
			if (sscanf (buf,"%s %s %s",dev,dir,stuff)>=2
				&& strcmp(dir,mpoint)==0){
				ret = true;
				break;
			}
		}
		fclose (fin);
	}
	return ret;
}

/*
	Mount a file system and return the result code
*/
static void audioserv_mount (
	int cli,
	const char *volume,
	int uid,
	bool silent)	// No mtab update
{
	if (!audioserv_ismounted(volume)){
		n_debug ("%s is not mounted\n",volume);
		if (fork()==0){
			char cmd[PATH_MAX];
			// Make the mount so
			// -the floppy is owned by the user triggering the mount
			// -very restrictive permissions
			// We assume that only DOS floppies will be mounted
			snprintf (cmd,sizeof(cmd)-1,"mount %s -o nosuid,nodev,uid=%d,umask=077 /mnt/%s"
				,silent ? "-n" : ""
				,uid,volume);
			REMAUDIO_CTL ctl;
			n_debug ("Executing %s\n",cmd);
			ctl.len = system (cmd);
			n_debug ("Produced %d -> is_mounted %d\n",ctl.len
				,audioserv_ismounted(volume));
			ctl.len = htonl(ctl.len);
			ctl.request = htonl(errno);
			ctl.p1 = 0;
			write (cli,&ctl,sizeof(ctl));
			_exit(0);
		}
	}
}		


int main (int argc,char *argv[])
{
	{
		FILE *fout = fopen ("/var/run/remaudioserv.pid","w");
		if (fout != NULL){
			fprintf (fout,"%d\n",getpid());
			fclose (fout);
		}
	}
	bool silent_mount = false;
	for (int i=1; i<argc; i++){
		const char *arg = argv[i];
		if (strcmp(arg,"--debug")==0){
			debug = true;
		}else if (strcmp(arg,"--silentmount")==0){
			silent_mount = true;
		}else{
			usage();
		}
	}
	CMDSOCK cmd (8010,1);
	signal (SIGCHLD,SIG_IGN);
	openlog ("remaudioserv",LOG_PID,LOG_DAEMON);
	struct VOLUME_CTL{
		const char *volume;
		time_t umount_time;	// When should we umount the volume.
	}tbvol[]={
		{"floppy",-1},
		{"fd1",-1},
		{"cdrom",-1},
	};
	while (1){
		if (cmd.listen (1000000) != -1){
			REMAUDIO_OPEN op;
			int nb;
			int cli;
			while ((nb=cmd.readnext(&op,sizeof(op),cli))!=-1){
				op.version = ntohs (op.version);
				op.cmd = ntohs (op.cmd);
				op.mode = ntohl (op.mode);
				// fprintf (stderr,"Recoit nb=%d version=%d cmd=%d\n",nb,op.version,op.cmd);
				if (nb == 0){
					n_debug ("cli %d a ferme la connexion\n",cli);
				}else if (nb != (int)sizeof(op)){
					syslog (LOG_ERR,"Connection invalide\n");
				}else if (op.version != REMAUDIO_VERSION){
					syslog (LOG_ERR,"Invalid protocol version: expect %d, receive %d\n"
						,REMAUDIO_VERSION,op.version);
				}else if (op.cmd == REMAUDIO_OPENFILE){
					n_debug ("Requete %d :%s: %d\n",nb,op.fname,op.mode);
					if (strcmp(op.fname,"/dev/audio")==0){
						audioserv_output (cli,op.fname,op.mode);
					}else if (strcmp(op.fname,"/dev/dsp")==0
						|| strcmp(op.fname,"/dev/dsp1")==0){
						audioserv_output (cli,op.fname,op.mode);
					}else if (strcmp(op.fname,"/dev/mixer")==0){
						audioserv_output (cli,op.fname,op.mode);
					}else if (strcmp(op.fname,"/dev/cdrom")==0){
						audioserv_output (cli,op.fname,op.mode);
					}else if (strcmp(op.fname,"/dev/fd0")==0
						|| strcmp(op.fname,"/dev/fd0H1440")==0
						|| strcmp(op.fname,"/dev/fd1")==0
						|| strcmp(op.fname,"/dev/fd1H1440")==0
						|| strcmp(op.fname,"/dev/lp0")==0
						|| strcmp(op.fname,"/dev/lp1")==0
						|| strcmp(op.fname,"/dev/lp2")==0){
						audioserv_output (cli,op.fname,op.mode);
					}else{
						syslog (LOG_ERR,"Requte inconnu: :%s:\n",op.fname);
					}
				}else if (op.cmd == REMAUDIO_MOUNT){
					const char *volume = op.fname;
					n_debug ("mounting %s\n",volume);
					audioserv_mount (cli,volume,op.mode,silent_mount);
					time_t t = time(NULL);
					for (int i=0; i<3; i++){
						if (strcmp(tbvol[i].volume,volume)==0){
							tbvol[i].umount_time = t+3;
							break;
						}
					}
				}
				cmd.closecli (cli);
			}
			// Do the umount if needed
			time_t t = time(NULL);
			for (int i=0; i<3; i++){
				if (tbvol[i].umount_time != -1
					&& tbvol[i].umount_time < t){
					const char *volume = tbvol[i].volume;
					time_t next_time = -1;
					if (audioserv_ismounted(volume)){
						char cmd[PATH_MAX];
						snprintf (cmd,sizeof(cmd)-1,"umount %s /mnt/%s"
							,silent_mount ? "-n" : "",volume);
						//printf ("%ld %s\n",t,cmd);
						if (system (cmd) != 0){
							// umount failed, probably busy, wait 3 more seconds
							next_time = t + 3;
						}
					}
					tbvol[i].umount_time = next_time;
				}
			}
		}
	}
}

