#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"
#include "litefs.p"
#include "remaudio.h"
#include "bufinput.h"

static void usage()
{
	fprintf (stderr,"litefs-proxy version 1.1\n"
		"\n"
		"litefs-proxy --socket sock_path [ --fork ] [ --debug] [ --exec command ...]\n"
		"litefs-proxy --pipe [ --fork] [ --debug] [ --exec command ...]\n"
		"litefs-proxy --connect host port [ --fork] [ --debug] [ --exec command ...]\n");
}

static bool debug=false;

void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}

// static int tbprog[1024];	// prog_handle -> client handle


/*
	Process message from the back-end.
	Return -1 if any error.

	At some point, we will introduce some optimisation to
	enhance the performance of litefs over high latency link.
	These enhancements include
		performing LITE_WRITE without waiting for an immediate
		answer (but report success to the caller). This
		will allow streaming of write request

		performs read ahread for LITE_READ.

	Those tricks will introduce asyncrhronous answers from
	the back-end and will be managed here.
*/
static int proxy_processopt (BUFINPUT_REPLY &buf_proxyin)
{
	int ret = buf_proxyin.fill() > 0 ? 0 : -1;
	int n;
	LITE_REPLY_DATA rep;
	while ((n=buf_proxyin.read(rep))>0){
		if (rep.status == LITE_PDATA_OUT
			|| rep.status == LITE_PDATA_ERR
			|| rep.status == LITE_STATUS_EXEC){
			n_debug ("process exec status=%d prog_handle %d\n",rep.status
				,rep.client_handle);
		}else{
			n_debug ("process other reply %d, client_handle=%d\n"
				,rep.status,rep.client_handle);
		}
		write (rep.client_handle,&rep,n);
	}
	return ret;
}


static int proxy_doreq (
	BUFINPUT_REQUEST &buf_client_in,
	int client_out,
	BUFINPUT_REPLY &buf_proxy_in,
	int proxy_out,
	bool no_shutdown)
{
	int ret = -1;
	LITE_REQUEST_DATA data;
	bool first = true;
	LITEFSD_CTX ctx;
	ctx.fd = client_out;
	ctx.client_handle = -1;
	while (1){
		int n;
		if (first){
			n_debug ("readblock start\n");
			n=buf_client_in.readblock (data);
			n_debug ("readblock end %d\n",n);
			first = false;
		}else{
			n_debug ("read no block start\n");
			n=buf_client_in.read (data);
			n_debug ("read no block end %d\n",n);
		}
		if (n > 0){
			data.client_handle = client_out;
			if (data.cmd == LITE_SHUTDOWN && no_shutdown){
				// We shutdown the proxy, but not the back-end
				litefsd_sendok (ctx);
				exit (0);
			}else if (data.cmd == LITE_EXEC){
				n_debug ("-proxy exec n=%d :%s:\n",n,data.buf);
				if (write (proxy_out,&data,n) == n){
					proxy_processopt (buf_proxy_in);
					ret = 0;
					#if 0
					LITE_REPLY_DATA rep;
					if ((n = buf_proxy_in.readblock (rep)) > 0){
						n_debug ("prog_handle = %d\n",rep.data);
						// tbprog[rep.data] = client_out;
						write (client_out,&rep,n);
						ret = 0;
					}else{
						n_debug ("exec fail\n");
						litefsd_sendfail (ctx);
					}
					#endif
				}
			}else if (data.cmd == LITE_EXECIN
				|| data.cmd == LITE_SIGNAL
				|| data.cmd == LITE_SETWIN){
				// No answer needed
				write (proxy_out,&data,n);
				ret = 0;
			}else{
				n_debug ("-proxy write n=%d cmd=%d\n",n,data.cmd);
				int writelen;
				if ((writelen=write (proxy_out,&data,n)) == n){
					proxy_processopt (buf_proxy_in);
					ret = 0;
					#if 0
						buf_proxy_in.fill();
						LITE_REPLY_DATA rep;
						while ((n = buf_proxy_in.read (rep)) > 0){
							n_debug ("-proxy read %d len %d\n",n,rep.len);
							if (write (client_out,&rep,n)==n){
								ret = 0;
							}
						}
						n_debug ("-proxy write end ret = %d %d(%s)\n"
							,ret,errno,strerror(errno));
					#endif
				}else{
					n_debug ("-proxy write n = %d != %d\n",n,writelen);
				}
			}
		}else{
			break;
		}
	}
	return ret;
}



static int proxy_fdset (fd_set &fdin, int maxfd, int handle)
{
	if (handle != -1){
		FD_SET (handle,&fdin);
		if (handle > maxfd) maxfd = handle;
	}
	return maxfd;
}


static void proxy_loop (
	int sock,
	int pipein,
	int pipeout,
	int proxyin,
	int proxyout,
	bool no_shutdown)		// Do not pass the shutdown signal to the backend
{
	BUFINPUT_REQUEST *tbfd[1024];
	BUFINPUT_REQUEST buf_pipein(pipein);
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = 0;
	maxfd = proxy_fdset (fdin,maxfd,sock);
	maxfd = proxy_fdset (fdin,maxfd,pipein);
	maxfd = proxy_fdset (fdin,maxfd,pipeout);
	maxfd = proxy_fdset (fdin,maxfd,proxyin);
	BUFINPUT_REPLY buf_proxyin (proxyin);
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		if (select (maxfd+1,&in,NULL,NULL,NULL)!=-1){
			int orig_nbfd = nbfd;
			if (sock != -1 && FD_ISSET (sock,&in)){
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_UNIX;
				int s = accept (sock,(struct sockaddr*)&unc,&len);
				tbfd[nbfd++] = new BUFINPUT_REQUEST(s);
				FD_SET (s,&fdin);
				if (s >= maxfd) maxfd = s + 1;
			}
			if (pipein != -1 && FD_ISSET(pipein,&in)){
				if (proxy_doreq (buf_pipein,pipeout,buf_proxyin,proxyout
					,no_shutdown) == -1){
					break;
				}
			}
			if (FD_ISSET(proxyin,&in)){
				if (proxy_processopt(buf_proxyin)==-1){
					break;
				}
			}
			bool redo_fdin = false;
			for (int i=0; i< orig_nbfd; i++){
				BUFINPUT_REQUEST *ptfd = tbfd[i];
				int fd = ptfd->fd;
				if (FD_ISSET(fd,&in)){
					if (proxy_doreq (*ptfd,fd,buf_proxyin,proxyout
						,no_shutdown) == -1){
						close (ptfd->fd);
						delete ptfd;
						redo_fdin = true;
						tbfd[i] = NULL;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = 0;
				maxfd = proxy_fdset (fdin,maxfd,sock);
				maxfd = proxy_fdset (fdin,maxfd,pipein);
				maxfd = proxy_fdset (fdin,maxfd,proxyin);
				int copyfd = 0;
				for (int i=0; i<nbfd; i++){
					if (tbfd[i] != NULL){
						maxfd = proxy_fdset (fdin,maxfd,tbfd[i]->fd);
						tbfd[copyfd++] = tbfd[i];
					}
				}
				nbfd = copyfd;
			}
		}
	}
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
static int litefs_receive(int fd)
{
	int ret = -1;
	LITE_REPLY rep;
	int len = read (fd,&rep,sizeof(rep));
	if (len == sizeof(LITE_REPLY)){
		if (rep.status == LITE_STATUS_OK || rep.status == LITE_STATUS_SEEK){
			ret = rep.data;
			errno = 0;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}


static void litefs_sendreq (int fd, LITE_REQUEST &req, int len)
{
	req.len = len;
	int ret = write (fd,&req,sizeof(LITE_REQUEST)+len);
	if (ret==-1){
		perror ("sendmsg");
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		int sock = -1;
		int pipein = -1;
		int pipeout = -1;
		int proxyin = -1;
		int proxyout = -1;
		bool dofork = false;
		bool no_shutdown = false;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				sock = litefsd_server(argv[i]);
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				setsid();	// Make sure that we are not killed when
							// the session leader die
				signal (SIGHUP,SIG_IGN);
				dofork = true;
			}else if (strcmp(argv[i],"--pipe")==0){
				pipein = 0;
				pipeout = 1;
			}else if (strcmp(argv[i],"--connect")==0){
				i++;
				const char *host = argv[i];
				i++;
				const char *port = argv[i];
				int handle = -1;
				if (isdigit(*port)){
					handle = cmdsock_connect (host,atoi(port),3);
				}else{
					handle = cmdsock_connect (host,port,3);
				}
				if (handle != -1){
					proxyin = proxyout = handle;
					no_shutdown = true;
				}
			}else if (strcmp(argv[i],"--exec")==0){
				if (i == argc -1){
					usage();
					exit (-1);
				}else{
					int tbin[2],tbout[2];
					if (pipe(tbin)==-1 || pipe(tbout)==-1){
						fprintf (stderr,"Can't setup pipe (%s)\n",strerror(errno));
					}else{
						int pid = fork();
						if (pid == -1){
							fprintf (stderr,"Can't fork new process (%s)\n"
								,strerror(errno));
						}else if (pid == 0){
							close (tbout[1]);
							close (tbin[0]);
							dup2 (tbin[1],1);
							dup2 (tbout[0],0);
							close (tbout[0]);
							close (tbin[1]);
							char line[2000];
							line[0] = '\0';
							for (int j=i+1; j<argc; j++){
								strcat (line,argv[j]);
								strcat (line," ");
							}
							system (line);
							_exit (0);
						}else{
							close (tbout[0]);
							close (tbin[1]);
							proxyin = tbin[0];
							proxyout = tbout[1];
						}
						i=argc-1;
					}
				}
			}else{
				usage();
				exit (-1);
			}
		}
		if ((sock != -1 || pipein != -1) && proxyin != -1){
			if (dofork){
				// First, make sure the link is established
				// before forking. This allows ssh to request
				// a password
				LITE_REQUEST req;
				req.cmd = LITE_VERSION;
				req.mode = LITEFS_PROTOCOL_VERSION;
				litefs_sendreq (proxyout,req,0);
				ret = litefs_receive(proxyin);
				if (ret == -1){
					fprintf (stderr,"Incompatible litefs backend\n");
					return -1;
				}
				if (fork() != 0) exit(0);
			}
			proxy_loop(sock,pipein,pipeout,proxyin,proxyout,no_shutdown);
		}else{
			usage();
		}
	}
	return ret;
}




