#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <pty.h>
#include <sys/un.h>
#include "litefs.h"
#include "litefs.p"
#include "bufinput.h"

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static void usage()
{
	fprintf (stderr,"litefsd version 1.0\n"
		"\n"
		"litefsd --pipe [ common options]\n"
		"litefsd --socket sock_path [--socket sock_path ] [ common options ]\n"
		"litefsd --tcpport port [--tcpport port ] [ common options ]\n"
		"\n"
		"common options: --chdir path\n"
		"                --debug\n"
		"                --fork\n"
		"                --pidfile path\n");
}

static const char *workdir;
static int len_workdir;

struct ONEMOUNT{
	const char *path;
	int lenpath;
	bool is_mounted;
	time_t may_umount;	// At which time it may be umounted
};

class TBMOUNT{
public:
	bool silentmount;	// Use option -n
	ONEMOUNT tb[100];
	int nbmount;
	TBMOUNT(){
		nbmount = 0;
	}
	/*~PROTOBEG~ TBMOUNT */
public:
	void add (const char *path);
	void check_mount (const char *path);
	void check_umount (void);
	/*~PROTOEND~ TBMOUNT */
};

PUBLIC void TBMOUNT::add (const char *path)
{
	tb[nbmount].may_umount = 0;
	tb[nbmount].is_mounted = false;
	tb[nbmount].lenpath = strlen(path);
	tb[nbmount++].path = path;
}

PUBLIC void TBMOUNT::check_umount ()
{
	time_t t = time(NULL);
	ONEMOUNT *pt = tb;
	for (int i=0; i<nbmount; i++,pt++){
		if (pt->is_mounted && pt->may_umount <= t){
			char cmd[PATH_MAX];
			snprintf (cmd,sizeof(cmd)-1,"umount %s %s"
				,silentmount ? "-n" : ""
				,pt->path);
			n_debug ("executing %s\n",cmd);
			if (system(cmd)==0){
				pt->is_mounted = false;
			}else{
				pt->may_umount = t+3;
			}
		}
	}
}
PUBLIC void TBMOUNT::check_mount (const char *path)
{
	time_t t = time(NULL);
	ONEMOUNT *pt = tb;
	for (int i=0; i<nbmount; i++,pt++){
		int lenpath = pt->lenpath;
		if (strncmp(path,pt->path,lenpath)==0
			&& (path[lenpath] == '\0' || path[lenpath] == '/')){
			if (!pt->is_mounted){
				char cmd[PATH_MAX];
				snprintf (cmd,sizeof(cmd)-1,"mount %s %s"
					,silentmount ? "-n" : ""
					,pt->path);
				n_debug ("executing %s\n",cmd);
				if (system(cmd)==0){
					pt->is_mounted = true;
				}
			}
			pt->may_umount = t+2;
			break;
		}
	}
}


static void litefsd_2abs(
	const char *arg,
	char abspath[PATH_MAX])
{
	if (arg[0] == '/' && arg[1] == '/'){
		strncpy (abspath,arg,PATH_MAX);
	}else{
		strcpy (abspath,workdir);
		strncpy (abspath+len_workdir,arg,PATH_MAX-len_workdir);
	}
	abspath[PATH_MAX-1] = '\0';
}


/*
	Convert the request to an absolute path by matching a volume
	definition. Send failure to the client if there is no matching
	volume or the request is invalid (.. in path)

	Return -1 if this is not a valid request.
*/
static int litefsd_translate (
	LITEFSD_CTX &ctx,
	const char *path,
	char abspath[PATH_MAX],
	TBMOUNT &tbauto)
{
	/* #Specification: litefs server / mapping the working dir
		litefs operates in the current directory unless it is forced
		to operate in the root. The users sees a litefs volume as
		a sub-directory (~/n/volume), so when it access a file
		~/n/volume/file is translate on the litefsd side to
		current_dir/file

		If the user specify a path like this: ~/n/volume//file, then
		this is converted to /file on the server side. This was done
		because it is expect that litefs will generally use to access
		a specific hierarchy on the server side, so by using
		the working directory of litefsd, we get this result.
	*/
	litefsd_2abs (path,abspath);
	tbauto.check_mount(abspath);
	return 0;
}

static int litefsd_translate (
	LITEFSD_CTX &ctx,
	LITE_REQUEST_DATA &req,
	char abspath[PATH_MAX],
	TBMOUNT &tbauto)
{
	int ret = -1;
	n_debug ("translate len = %d\n",req.len);
	if (req.len > 0){
		ret = litefsd_translate (ctx,req.buf,abspath,tbauto);
	}else{
		litefsd_sendfail (ctx);
	}
	return ret;
}

class PROGCTL{
public:
	int client;		// Handle to talk to the client
	int client_handle;	// Received with the first request
	int in,out,err;	// Handles to talk/listen to the program
					// as seen by the program
					// ("in" is used to write to the program standard input)
	pid_t pid;
	/*~PROTOBEG~ PROGCTL */
public:
	PROGCTL (void);
	void closeprog (void);
	/*~PROTOEND~ PROGCTL */
};

PUBLIC PROGCTL::PROGCTL()
{
	client = -1;
	pid = -1;
	in = out = err = -1;
}
/*
	Close the file handle used to talk to the program
*/
PUBLIC void PROGCTL::closeprog()
{
	close (in);
	close (out);
	close (err);
	in = out = err = -1;
}

/*
	Parse a bunch of sequential null terminated string from a buffer
*/
static const char *litefsd_pickstrs(
	const char *buf,
	const char *str[],
	int nbstr)
{
	for (int i=0; i<nbstr; i++){
		str[i] = buf;
		while (*buf != '\0') buf++;
		buf++;
	}
	return buf;
}

static int litefsd_setupexec (
	const char *abspath,
	PROGCTL &ctl,
	LITE_EXECINFO_DATA &info)
{
	int ret = -1;
	ctl.in = ctl.out = ctl.err = -1;
	int tbin[2],tbout[2],tberr[2];
	tbin[0] = -1;
	tbout[0] = -1;
	tberr[0] = -1;
	if (info.win.height != 0){
		int master,slave;
		struct winsize wins;
		memset (&wins,0,sizeof(wins));
		wins.ws_row = info.win.height;
		wins.ws_col = info.win.width;
		if (openpty(&master,&slave,NULL,NULL,&wins)!=-1){
			tbout[1] = slave;
			tbout[0] = master;
			if (info.isatty0){
				tbin[1] = dup(master);
				tbin[0] = dup(slave);
			}else{
				pipe(tbin);
			}
		}
	}else{
		pipe (tbin);
		pipe (tbout);
	}
	if (tbin[0] != -1
		&& tbout[0] != -1
		&& pipe (tberr) != -1){
		pid_t pid = fork();
		if (pid == 0){
			dup2 (tbin[0],0);
			dup2 (tbout[1],1);
			dup2 (tberr[1],2);
			for (int i=3; i<1024; i++) close (i);
			const char *argv[info.nbarg+1];
			const char *pt = litefsd_pickstrs (info.data,argv,info.nbarg);
			argv[info.nbarg] = NULL;
			// We merge the environment variable received with the
			// one we currently have
			const char *env[info.nbenv+1];
			litefsd_pickstrs (pt,env,info.nbenv);
			for (int i=0; i<info.nbenv; i++) putenv (env[i]);
			execv (abspath,(char**)argv);
			_exit (-1);
		}else if (pid == -1){
			close (tbin[0]);
			close (tbin[1]);
			close (tbout[0]);
			close (tbout[1]);
			close (tberr[0]);
			close (tberr[1]);
		}else{
			close (tbin[0]);
			close (tbout[1]);
			close (tberr[1]);
			ctl.in = tbin[1];
			ctl.out = tbout[0];
			ctl.err = tberr[0];
			ctl.pid = pid;
			ret = 0;
		}
	}else{
		close (tbin[0]); close (tbin[1]);
		close (tbout[0]); close (tbout[1]);
		close (tberr[0]); close (tberr[1]);
	}
	return ret;
}

class  PROGCTLS{
	int first;
	int maxctls;
	PROGCTL *tb;
	/*~PROTOBEG~ PROGCTLS */
public:
	PROGCTLS (void);
	PROGCTL *alloc (int fd, int client_handle);
	int fdset (fd_set&fdin, int maxfd);
	void process (fd_set&fdin, bool&redoset);
	void release (PROGCTL *ctl);
	void sendin (int client_handle, void *buf, int len);
	void sendsignal (int client_handle, int sig);
	int sendstatus (pid_t pid, int status);
	void setwin (int client_handle, LITE_WININFO *win);
	/*~PROTOEND~ PROGCTLS */
};

PUBLIC PROGCTLS::PROGCTLS()
{
	first = 0;
	maxctls = 0;
	tb = NULL;
}

/*
	Return the first available PROGCTL
*/
PUBLIC PROGCTL *PROGCTLS::alloc (int fd, int client_handle)
{
	if (first == maxctls){
		int old = maxctls;
		maxctls += 10;
		tb = (PROGCTL *)realloc (tb,maxctls*sizeof(PROGCTL));
		PROGCTL *ctl = tb+old;
		for (int i=old; i<maxctls; i++, ctl++){
			ctl->client = i+1;
			ctl->in = ctl->out = ctl->err = -1;
			ctl->client_handle = -1;
		}
	}
	PROGCTL *ret = tb + first;
	ret->client = fd;
	ret->client_handle = client_handle;
	first = ret->client;
	return ret;
}

PUBLIC void PROGCTLS::release (PROGCTL *ctl)
{
	ctl->client = first;
	ctl->closeprog();
	ctl->client_handle = -1;
	first = (int)(ctl - tb);
}

PUBLIC int PROGCTLS::fdset (fd_set &fdin, int maxfd)
{
	PROGCTL *pt = tb;
	for (int i=0; i<maxctls; i++, pt++){
		if (pt->out != -1){
			FD_SET (pt->out,&fdin);
			if (pt->out >= maxfd) maxfd = pt->out + 1;
			FD_SET (pt->err,&fdin);
			if (pt->err >= maxfd) maxfd = pt->err + 1;
		}
	}
	return maxfd;
}

PUBLIC int PROGCTLS::sendstatus (pid_t pid, int status)
{
	int ret = -1;
	PROGCTL *pt = tb;
	for (int i=0; i<maxctls; i++, pt++){
		if (pt->pid == pid){
			// We process the status only if the connection
			// was closed with the process. Sendstatus will be called
			// again
			if (pt->out == -1){
				LITE_REPLY rep;
				rep.status = LITE_STATUS_EXEC;
				rep.data = i;
				rep.errcode = status;
				rep.client_handle = pt->client_handle;
				rep.len = 0;
				write (pt->client,&rep,sizeof(rep));
				release (pt);
				ret = 0;
			}
			break;
		}
	}
	return ret;
}

const int MAXSTATUS=10;
static struct TBSTATUS{
	pid_t pid;
	int status;
}tbstatus[MAXSTATUS];

static void fct_child (int)
{
	pid_t pid;
	int status;
	pid = wait(&status);
	if (pid != -1){
		TBSTATUS *ptt = tbstatus;
		for (int i=0; i<MAXSTATUS; i++,ptt++){
			if (ptt->pid == 0){
				ptt->pid = pid;
				ptt->status = status;
				break;
			}
		}
	}
}

/*
	Read the stdout or error channel from a running program and
	send this to the client.
*/
static int litefsd_sendpdata (
	int client_handle,
	int fd,
	LITE_STATUS status,
	int client)
{
	LITE_REPLY_DATA rep;
	rep.status = status;
	rep.client_handle = client_handle;
	rep.errcode = 0;
	int len = read (fd,rep.buf,sizeof(rep.buf));
	n_debug ("litefsd_sendpdata: len=%d client_handle %d\n",rep.len,client_handle);
	if (len > 0){
		rep.len = len;
		write (client,&rep,sizeof(LITE_REPLY)+rep.len);
	}
	return len;
}

PUBLIC void PROGCTLS::process (fd_set &fdin, bool &redoset)
{
	PROGCTL *pt = tb;
	for (int i=0; i<maxctls; i++, pt++){
		if (pt->out != -1){
			bool doclose = false;
			if (FD_ISSET(pt->out,&fdin)){
				if (litefsd_sendpdata (pt->client_handle,pt->out,LITE_PDATA_OUT
					,pt->client) <=0){
					doclose = true;
				}
			}
			if (FD_ISSET(pt->err,&fdin)){
				if(litefsd_sendpdata (pt->client_handle,pt->err,LITE_PDATA_ERR
					,pt->client)<=0){
					doclose = true;
				}
			}
			if (doclose){
				redoset = true;
				pt->closeprog();
			}
		}
	}
}

PUBLIC void PROGCTLS::sendin (int client_handle, void *buf, int len)
{
	PROGCTL *p = tb;
	for (int i=0; i<maxctls; i++, p++){
		if (p->client_handle == client_handle){
			n_debug ("sendin len=%d pid=%d client_handle=%d\n"
					,len,p->pid,p->client_handle);
			if (len == 0){
				// EOF on stdin
				if (p->in == p->out){
					// Running in a PTY
					// Sending a hangup signal hum? Not sure about this one
					//kill (p->pid,SIGHUP);
				}
				close (p->in);
				p->in = -1;
			}else{
				write (p->in,buf,len);
			}
			break;
		}
	}
}
PUBLIC void PROGCTLS::sendsignal (int client_handle, int sig)
{
	PROGCTL *p = tb;
	for (int i=0; i<maxctls; i++, p++){
		if (p->client_handle == client_handle){
			kill (p->pid,sig);
			break;
		}
	}
}
PUBLIC void PROGCTLS::setwin (int client_handle, LITE_WININFO *win)
{
	PROGCTL *p = tb;
	for (int i=0; i<maxctls; i++, p++){
		if (p->client_handle == client_handle){
			struct winsize wins;
			if (ioctl(p->out,TIOCGWINSZ,&wins) != -1){
				wins.ws_col = win->width;
				wins.ws_row = win->height;
				ioctl(p->out,TIOCSWINSZ,&wins);
			}
			break;
		}
	}
}
/*
	Read a request from one client
*/
static int litefsd_doreq (
	BUFINPUT_REQUEST &buf_in,
	int fout,
	PROGCTLS &ctls,
	bool &redoset,		// Redo the set for select because ctls
						// have changes
	TBMOUNT &tbauto,
	bool &shutdown)
{
	const int maxdir = 1024;
	static DIR *tbdir[maxdir];
	LITE_REQUEST_DATA req;
	int ret = buf_in.fill() > 0 ? 0 : -1;
	int len;
	while ((len = buf_in.read(req))>0){
		n_debug ("req.cmd = %d len=%d\n",req.cmd,len);
		LITEFSD_CTX ctx;
		ctx.client_handle = req.client_handle;
		if (ctx.client_handle == -1) ctx.client_handle = fout;
		ctx.fd = fout;
		if (req.cmd == LITE_VERSION){
			if (req.mode != LITEFS_PROTOCOL_VERSION){
				litefsd_sendfail (ctx);
			}else{
				litefsd_sendok(ctx);
			}
		}else if (req.cmd == LITE_SHUTDOWN){
			litefsd_sendok(ctx);
			shutdown = true;
		}else if (req.cmd == LITE_OPEN){
			n_debug ("open :%s: %d %d\n",req.buf,req.mode,req.perm);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				int handle = open (abspath,req.mode,req.perm);
				n_debug ("open abspath :%s: handle %d\n",abspath,handle);
				if (handle != -1){
					litefsd_sendfd (ctx,handle);
				}else{
					litefsd_senderrno(ctx);
				}
			}
		}else if (req.cmd == LITE_CLOSE){
			n_debug ("close %d\n",req.handle);
			if (close(req.handle)!=-1){
				litefsd_sendok(ctx);
			}else{
				litefsd_senderrno(ctx);
			}
		}else if (req.cmd == LITE_SEEK){
			n_debug ("seek %d %d %d\n",req.handle,req.mode,req.perm);
			off_t off = lseek(req.handle,req.mode,req.perm);
			if (off != (off_t)-1){
				litefsd_sendseek(ctx,off);
			}else{
				litefsd_senderrno(ctx);
			}
		}else if (req.cmd == LITE_READ){
			n_debug ("read %d %d\n",req.handle,req.mode);
			LITE_REPLY_DATA rep;
			len = read (req.handle,rep.buf,req.mode);
			if (len >= 0){
				rep.len = len;
				rep.status = LITE_STATUS_OK;
				rep.client_handle = ctx.client_handle;
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}else{
				litefsd_senderrno(ctx);
			}
		}else if (req.cmd == LITE_WRITE){
			int len = write (req.handle,req.buf,req.len);
			if (len >= 0){
				litefsd_sendok(ctx,len);
			}else{
				litefsd_senderrno(ctx);
			}
		}else if (req.cmd == LITE_STAT){
			n_debug ("stat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("stat abspath :%s:\n",abspath);
				struct stat st;
				if (stat(abspath,&st)==-1){
					litefsd_senderrno(ctx);
				}else{
					litefsd_sendstat (ctx,st);
				}
			}
		}else if (req.cmd == LITE_LSTAT){
			n_debug ("lstat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("lstat abspath :%s:\n",abspath);
				struct stat st;
				if (lstat(abspath,&st)==-1){
					litefsd_senderrno(ctx);
				}else{
					litefsd_sendstat (ctx,st);
				}
			}
		}else if (req.cmd == LITE_FSTAT){
			struct stat st;
			if (fstat(req.handle,&st)==-1){
				n_debug ("fstat %d failed %d\n",req.handle,errno);
				litefsd_senderrno(ctx);
			}else{
				litefsd_sendstat (ctx,st);
				n_debug ("fstat %d ok\n",req.handle);
			}
		}else if (req.cmd == LITE_UNLINK){
			n_debug ("unlink :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("unlink abspath :%s:\n",abspath);
				if (unlink(abspath)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_RENAME){
			const char *newp = req.buf;
			while (*newp != '\0') newp++;
			newp++;
			n_debug ("rename :%s: :%s:\n",req.buf,newp);
			char old_abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,old_abspath,tbauto)!=-1){
				char new_abspath[PATH_MAX];
				litefsd_2abs (newp,new_abspath);
				n_debug ("rename abspath :%s: :%s:\n",old_abspath,new_abspath);
				if (rename(old_abspath,new_abspath)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_LINK){
			const char *newp = req.buf;
			while (*newp != '\0') newp++;
			newp++;
			n_debug ("link :%s: :%s:\n",req.buf,newp);
			char old_abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,old_abspath,tbauto)!=-1){
				char new_abspath[PATH_MAX];
				litefsd_2abs (newp,new_abspath);
				n_debug ("link abspath :%s: :%s:\n",old_abspath,new_abspath);
				if (link(old_abspath,new_abspath)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_SYMLINK){
			const char *oldp = req.buf;
			while (*oldp != '\0') oldp++;
			oldp++;
			n_debug ("link :%s: :%s:\n",oldp,req.buf);
			char new_abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,new_abspath,tbauto)!=-1){
				n_debug ("symlink abspath :%s: :%s:\n",oldp,new_abspath);
				if (symlink(oldp,new_abspath)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_UTIMES){
			n_debug ("utimes :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("utimes abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				int ok = -1;
				if (req.mode == -1 && req.perm == -1){
					ok = utimes (abspath,NULL);
				}else{
					struct timeval tvp[2];
					tvp[0].tv_sec = req.mode;
					tvp[1].tv_sec = req.perm;
					tvp[0].tv_usec = 0;
					tvp[1].tv_usec = 0;
					ok = utimes (abspath,tvp);
				}
				if (ok!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_ACCESS){
			n_debug ("access :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("access abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				int ok = access (abspath,req.mode);
				if (ok!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_RMDIR){
			n_debug ("rmdir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("rmdir abspath :%s:\n",abspath);
				if (rmdir(abspath)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_MKDIR){
			n_debug ("mkdir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("mkdir abspath :%s:\n",abspath);
				if (mkdir(abspath,req.mode)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_CHDIR){
			// chdir means, may chdir
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("chdir abspath :%s:\n",abspath);
				struct stat st;
				if (stat(abspath,&st)!=-1 && S_ISDIR(st.st_mode)){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_CHMOD){
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("chmod abspath :%s: %o\n",abspath,req.mode);
				if (chmod (abspath,req.mode)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_FCHMOD){
			n_debug ("fchmod %d\n",req.handle);
			if (fchmod(req.handle,req.mode)!=-1){
				litefsd_sendok (ctx);
			}else{
				litefsd_senderrno (ctx);
			}
		}else if (req.cmd == LITE_FCHOWN){
			n_debug ("fchown %d\n",req.handle);
			if (fchown(req.handle,req.mode,req.perm)!=-1){
				litefsd_sendok (ctx);
			}else{
				litefsd_senderrno (ctx);
			}
		}else if (req.cmd == LITE_CHOWN){
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("chown abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				if (chown (abspath,req.mode,req.perm)!=-1){
					litefsd_sendok (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_OPENDIR){
			n_debug ("opendir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("opendir abspath :%s:\n",abspath);
				DIR *dir = opendir (abspath);
				if (dir != NULL){
					int i;
					for (i=0; i<maxdir; i++){
						if (tbdir[i] == NULL){
							tbdir[i] = dir;
							litefsd_sendhandle (ctx,i);
							break;
						}
					}
					if (i==maxdir) litefsd_sendfail (ctx);
				}else{
					litefsd_senderrno (ctx);
				}
			}
		}else if (req.cmd == LITE_READDIR){
			n_debug ("readdir %d\n",req.handle);
			DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (ctx);
			}else{
				struct dirent *ent = readdir (dir);
				LITE_REPLY_DATA rep;
				rep.status = LITE_STATUS_OK;
				if (ent == NULL){
					rep.len = 0;
				}else{
					rep.len = sizeof(struct dirent);
					memcpy (rep.buf,ent,sizeof(struct dirent));
				}
				rep.client_handle = ctx.client_handle;
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}
		}else if (req.cmd == LITE_CLOSEDIR){
			n_debug ("closedir %d\n",req.handle);
			DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (ctx);
			}else if (closedir(dir)!=-1){
				litefsd_sendok(ctx);
			}else{
				litefsd_senderrno(ctx);
			}
		}else if (req.cmd == LITE_READLINK){
			n_debug ("readlink %s\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (ctx,req,abspath,tbauto)!=-1){
				n_debug ("readlink abspath :%s:\n",abspath);
				LITE_REPLY_DATA rep;
				len = readlink (abspath,rep.buf,sizeof(rep.buf));
				if (len == -1){
					litefsd_senderrno (ctx);
				}else{
					rep.len = len;
					rep.status = LITE_STATUS_OK;
					rep.client_handle = ctx.client_handle;
					write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
				}
			}
		}else if (req.cmd == LITE_EXEC){
			n_debug ("exec\n");
			if (req.len <= sizeof(LITE_EXECINFO)){
				litefsd_sendfail (ctx);
			}else{
				LITE_EXECINFO_DATA *info = (LITE_EXECINFO_DATA*)req.buf;
				n_debug ("exec :%s:\n",info->data);
				char abspath[PATH_MAX];
				if (litefsd_translate (ctx,info->data,abspath,tbauto)!=-1){
					PROGCTL *ctl = ctls.alloc(fout,ctx.client_handle);
					if (litefsd_setupexec (abspath,*ctl,*info) != -1){
						n_debug ("setupexec ok\n");
						LITE_REPLY rep;
						rep.status = LITE_STATUS_OK;
						rep.data = 0;
						rep.len = 0;
						rep.client_handle = ctx.client_handle;
						write (fout,&rep,sizeof(LITE_REPLY));
						redoset = true;
					}else{
						n_debug ("setupexec fail\n");
						ctls.release (ctl);
						litefsd_sendfail (ctx);
					}
				}
			}
		}else if (req.cmd == LITE_EXECIN){
			ctls.sendin (ctx.client_handle,req.buf,req.len);
		}else if (req.cmd == LITE_SIGNAL){
			ctls.sendsignal (ctx.client_handle,req.mode);
		}else if (req.cmd == LITE_SETWIN){
			ctls.setwin (ctx.client_handle,(LITE_WININFO*)req.buf);
		}else{
			litefsd_sendfail(ctx);
			ret = -1;
		}
	}
	return ret;
}
		
struct ONESOCK{
	int handle;
	int family;
};

class TBSOCK{
public:
	ONESOCK tb[100];
	int nbsock;
	TBSOCK(){
		nbsock = 0;
	}
	/*~PROTOBEG~ TBSOCK */
public:
	void add (int family, int sock);
	int fdset (fd_set&fdin, int maxfd);
	/*~PROTOEND~ TBSOCK */
};

PUBLIC void TBSOCK::add (int family, int sock)
{
	tb[nbsock].family = family;
	tb[nbsock++].handle = sock;
}

/*
	Fill an fd_set with all the listening socket handle.
	Return the new maxfd
*/
PUBLIC int TBSOCK::fdset (fd_set &fdin, int maxfd)
{
	for (int i=0; i<nbsock; i++){
		int sock = tb[i].handle;
		FD_SET (sock,&fdin);
		if (sock >= maxfd) maxfd = sock + 1;
	}
	return maxfd;
}


/*
	Process all request and connections. Infinite loop
*/
static void litefsd_loop (
	TBSOCK &tbsock,		// Unix/Tcp socket to receive new connection
	TBMOUNT &tbauto,	// Directory to mount/umount as needed
	int pipein,			// Pre-connected client or proxy
	int pipeout)		// Pre-connected client or proxy
{
	BUFINPUT_REQUEST buf_pipein(pipein);
	BUFINPUT_REQUEST *tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = tbsock.fdset (fdin,0);
	if (pipein != -1){
		FD_SET (pipein,&fdin);
		if (pipein >= maxfd) maxfd = pipein+1;
	}
	PROGCTLS ctls;
	bool shutdown = false;
	while (!shutdown){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		struct timeval timeout;
		timeout.tv_sec = 2;
		timeout.tv_usec = 0;
		if (select (maxfd,&in,NULL,NULL,&timeout)!=-1){
			int orig_nbfd = nbfd;
			bool redo_fdin = false;
			for (int i=0; i<tbsock.nbsock; i++){
				int sock = tbsock.tb[i].handle;
				if (FD_ISSET (sock,&in)){
					struct sockaddr_un unc;
					size_t len = sizeof(unc);
					unc.sun_family = tbsock.tb[i].family;
					int s = accept (sock,(struct sockaddr*)&unc,&len);
					tbfd[nbfd++] = new BUFINPUT_REQUEST(s);
					FD_SET (s,&fdin);
					if (s >= maxfd) maxfd = s + 1;
				}
			}
			if (pipein != -1 && FD_ISSET(pipein,&in)){
				if (litefsd_doreq (buf_pipein,pipeout,ctls,redo_fdin
					,tbauto,shutdown)
					== -1){
					break;
				}
			}
			for (int i=0; i< orig_nbfd; i++){
				BUFINPUT_REQUEST *ptfd = tbfd[i];
				int fd = ptfd->fd;
				if (FD_ISSET(fd,&in)){
					if (litefsd_doreq (*ptfd,fd,ctls,redo_fdin
						,tbauto,shutdown) == -1){
						close (fd);
						redo_fdin = true;
						delete ptfd;
						tbfd[i] = NULL;
					}
				}
			}
			ctls.process (in,redo_fdin);
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = tbsock.fdset (fdin,0);
				if (pipein != -1){
					FD_SET (pipein,&fdin);
					if (pipein >= maxfd) maxfd = pipein+1;
				}
				int copyfd = 0;
				for (int i=0; i<nbfd; i++){
					if (tbfd[i] != NULL){
						int fd = tbfd[i]->fd;
						FD_SET (fd,&fdin);
						if (fd >= maxfd) maxfd = fd + 1;
						tbfd[copyfd++] = tbfd[i];
					}
				}
				nbfd = copyfd;
				maxfd = ctls.fdset (fdin,maxfd);
			}
		}
		tbauto.check_umount();
		TBSTATUS *ptt = tbstatus;
		for (int i=0; i<MAXSTATUS; i++,ptt++){
			if (ptt->pid != 0){
				if (ctls.sendstatus (ptt->pid,ptt->status) != -1){
					ptt->pid = 0;
				}
			}
		}
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		TBSOCK tbsock;		// Listening sockets
		TBMOUNT tbauto;		// Dir that should be automounted
		int pipein = -1;
		int pipeout = -1;
		const char *newdir = NULL;
		bool dofork = false;
		const char *pidfile = NULL;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				int sock = litefsd_server(argv[i]);
				if (sock != -1){
					tbsock.add (AF_UNIX,sock);
				}
			}else if (strcmp(argv[i],"--pipe")==0){
				pipein = 0;
				pipeout = 1;
			}else if (strcmp(argv[i],"--tcpport")==0){
				i++;
				int sock = litefsd_tcpserver(argv[i]);
				if (sock != -1){
					tbsock.add (AF_INET,sock);
				}
			}else if (strcmp(argv[i],"--automount")==0){
				i++;
				tbauto.add (argv[i]);
			}else if (strcmp(argv[i],"--chdir")==0){
				i++;
				newdir = argv[i];
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				dofork = true;
			}else if (strcmp(argv[i],"--pidfile")==0){
				i++;
				pidfile = argv[i];
			}else{
				usage();
				exit (-1);
			}
		}
		if (newdir != NULL && chdir(newdir)==-1){
			fprintf (stderr,"Can't change to directory %s (%s)\n"
				,newdir,strerror(errno));
			exit (-1);
		}
		char cwd[PATH_MAX];
		if (getcwd(cwd,PATH_MAX)==NULL){
			fprintf (stderr,"Can't locate current working directory (%s)\n"
				,strerror(errno));
		}else{
			workdir = cwd;
			len_workdir = strlen(cwd);
			if (tbsock.nbsock > 0 || pipein != -1){
				if (dofork){
					if (fork() != 0) exit(0);
					setsid();
				}
				if (pidfile != NULL){
					FILE *fout = fopen (pidfile,"w");
					if (fout != NULL){
						fprintf (fout,"%d\n",getpid());
						fclose (fout);
					}
				}
				signal (SIGCLD,fct_child);
				signal (SIGWINCH,SIG_IGN);
				signal (SIGHUP,SIG_IGN);
				litefsd_loop(tbsock,tbauto,pipein,pipeout);
			}else{
				usage();
			}
		}
	}
	return ret;
}



