/* #Specification: persmount_remote / principles
	This is a setuid program allowing normal user to do NFS mount
	of specific volume on remote host. The program receives
	as arguments the hostname and the volume. It will then do the following

	<SGML>
	<itemize>
	<item>Send a mount request to the audioserver running on that host.
		While the server may run an automounter, we have found that
		NFS server and autom0unter are not reliable.
	<item>Do an NFS mount. The volume on the server is always /mnt/vol
		where vol is the second argument. The mount point is always
		/mnt/hostname/vol, where hostname is the first argument.
	</itemize>
	</SGML>

	Because this is a setuid program, suitable validation is done
	for the hostname and volume.
*/
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include "remaudio.h"

static int persmount_mkdiranc (const char *path)
{
	int ret = mkdir (path,0755);
	if (ret == -1 && errno != EEXIST){
		char tmp[strlen(path)+1];
		strcpy (tmp,path);
		char *pt = strrchr(tmp,'/');
		if (pt != NULL){
			*pt = '\0';
			if (persmount_mkdiranc (tmp) != -1) ret = mkdir (path,0755);
		}
	}
	return ret;
}

static bool persmount_validvolume (const char *volume)
{
	bool ret = true;
	if (!isalpha(*volume)){
		ret = false;
	}else{
		while (*volume != '\0'){
			if (!isdigit(*volume) && !isalpha(*volume)){
				ret = false;
				break;
			}
			volume++;
		}
	}
	return ret;
}

static bool persmount_validhost(const char *host)
{
	// A host name may only contain digits, letters, dots and dashes.
	// It can contain a sequence of two dots.
	bool ret = true;
	if (strstr (host,"..")!=NULL){
		ret = false;
	}else{
		while (*host != '\0'){
			if (!isalpha(*host)
				&& !isdigit(*host)
				&& *host != '-'
				&& *host != '.'){
				ret = false;
				break;
			}
			host++;
		}
	}
	return ret;
}

static int remaudio_bufread (int cli, void *buf, int len)
{
	char *pbuf = (char*)buf;
	int sofar = 0;
	while (sofar < len){
		int nb=read(cli,pbuf+sofar,len-sofar);
		if (nb > 0){
			sofar += nb;
		}else if (errno != EAGAIN){
			break;
		}
	}
	return sofar != len ? -1 : len;
}



static int remaudio_readack(int fd, REMAUDIO_CTL &ctl)
{
	int ret = -1;
	if (remaudio_bufread (fd,&ctl,sizeof(REMAUDIO_CTL))!=-1){
		//printf ("tmp = %d %d\n",ctl.len, ctl.p1);
		ret = ctl.len;
		errno = ctl.request;
	}
	return ret;
}


static int remaudio_readack(int fd)
{
	REMAUDIO_CTL ctl;
	return remaudio_readack (fd,ctl);
}

static int remaudio_connect (
	const char *host,
	const char *mpoint)
{
	int ret = -1;
	int fd = cmdsock_connect (host,8010,1);
	if (fd != -1){
		REMAUDIO_OPEN op;
		op.version = htons(REMAUDIO_VERSION);
		op.cmd = htons(REMAUDIO_MOUNT);
		strcpy (op.fname,mpoint);
		op.mode = 0;
		write (fd,&op,sizeof(op));
		ret = remaudio_readack(fd);
		close (fd);
	}
	return ret;
}



int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc != 3){
		fprintf (stderr,"persmount_remote host volume\n");
	}else if (!persmount_validhost(argv[1])){
		fprintf (stderr,"Invalid host name\n");
	}else if (!persmount_validvolume (argv[2])){
		fprintf (stderr,"Invalid volume\n");
	}else{
		setgid (0);
		char local_dir[PATH_MAX];
		snprintf (local_dir,sizeof(local_dir)-1,"/mnt/%s/%s",argv[1],argv[2]);
		persmount_mkdiranc (local_dir);
		char remote_dir[PATH_MAX];
		snprintf (remote_dir,sizeof(remote_dir)-1,"/mnt/%s",argv[2]);
		if (remaudio_connect(argv[1],argv[2])==0){
			char cmd[PATH_MAX];
			snprintf (cmd,sizeof(cmd)-1,"/bin/mount %s:%s %s",argv[1]
				,remote_dir,local_dir);
			setuid (0);
			ret = system (cmd);
		}else{
			fprintf (stderr,"remote mount failed\n");
		}
	}
	return ret;
}

