#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include <netinet/in.h>
#include <linux/soundcard.h>
#include <linux/cdrom.h>
#include "remaudio.h"
#include "dispatch.h"

static bool debug=false;

void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vprintf (ctl,list);
		va_end (list);
	}
}



enum REMAUDIO_TYPE {
	NET_NOSPECIAL=0,
	NET_AUDIO,
	NET_DSP,
	NET_MIXER,
	NET_FLOPPY,
	NET_CDROM,
};

static const int MAX_FILE_TB=1024;
static REMAUDIO_TYPE tb[MAX_FILE_TB];

static int remaudio_bufread (int cli, void *buf, int len)
{
	char *pbuf = (char*)buf;
	int sofar = 0;
	while (sofar < len){
		int nb=__libc_read(cli,pbuf+sofar,len-sofar);
		if (nb > 0){
			sofar += nb;
		}else if (errno != EAGAIN){
			break;
		}
	}
	return sofar != len ? -1 : len;
}



static int remaudio_readack(int fd, REMAUDIO_CTL &ctl)
{
	int ret = -1;
	if (remaudio_bufread (fd,&ctl,sizeof(REMAUDIO_CTL))!=-1){
		//printf ("tmp = %d %d\n",ctl.len, ctl.p1);
		ret = ntohl(ctl.len);
		errno = ntohl(ctl.request);
	}
	return ret;
}
static int remaudio_readack(int fd)
{
	REMAUDIO_CTL ctl;
	return remaudio_readack (fd,ctl);
}

char __audioserver[200];	// Name of the audioserver

static int remaudio_write (int fd, REMAUDIO_OPEN &op)
{
	op.version = htons(op.version);
	op.cmd = htons (op.cmd);
	op.mode = htonl (op.mode);
	return __libc_write (fd,&op,sizeof(REMAUDIO_OPEN));
}

static int remaudio_write (int fd, REMAUDIO_CTL &ctl)
{
	ctl.cmd = htons (ctl.cmd);
	ctl.request = htons (ctl.request);
	ctl.len = htonl  (ctl.len);
	ctl.p1  = htonl  (ctl.p1);
	ctl.p2  = htonl  (ctl.p2);
	ctl.p3  = htonl  (ctl.p3);
	return __libc_write (fd,&ctl,sizeof(REMAUDIO_CTL));
}



static int remaudio_connect (
	const char *fname,
	int mode,
	REMAUDIO_TYPE type)
{
	n_debug ("Connect to :%s: for file %s\n",__audioserver,fname);
	int fd = __clientsock_connect (__audioserver,8010,1);
	//printf ("Connect to :%s: -> %d\n",__audioserver,fd);
	if (fd != -1){
		REMAUDIO_OPEN op;
		op.version = REMAUDIO_VERSION;
		op.cmd = REMAUDIO_OPENFILE;
		strcpy (op.fname,fname);
		op.mode = mode;
		remaudio_write (fd,op);
		remaudio_readack(fd);
		tb[fd] = type;
	}
	return fd;
}
enum OPEN_STATE {
	OPEN_UNKNOWN,
	OPEN_LOCAL,
	OPEN_REMOTE
};
static OPEN_STATE state = OPEN_UNKNOWN;

static void remaudio_setstate()
{
	// We check once if we are running in remote ressource mode
	state = OPEN_LOCAL;
	const char *s = getenv ("REMAUDIOHOST");
	if (s == NULL) s = getenv("DISPLAY");
	if (s != NULL){
		strcpy (__audioserver,s);
		char *pt = strchr(__audioserver,':');
		if (pt != NULL) *pt = '\0';
		if (__audioserver[0] != '\0'){
			char hname[PATH_MAX];
			if (gethostname (hname,PATH_MAX-1)!=-1
				&& strcmp(__audioserver,hname)!=0){
				state = OPEN_REMOTE;
			}
		}else{
			strcpy (__audioserver,"127.0.0.1");
		}
	}else{
		strcpy (__audioserver,"127.0.0.1");
	}
}

/*
	Return true if some devices are remote
*/
bool __remaudio_isremote()
{
	if (state == OPEN_UNKNOWN){
		remaudio_setstate();
	}
	return state == OPEN_REMOTE;
}

static int remaudio_open (const char *fname, int mode, int perm)
{
	int ret = DISPATCH_DONTCARE;
	if (state == OPEN_UNKNOWN){
		remaudio_setstate();
	}
	if (state == OPEN_REMOTE){
		if (strcmp(fname,"/dev/audio")==0){
			ret = remaudio_connect (fname,mode,NET_AUDIO);
		}else if (strcmp(fname,"/dev/dsp")==0
			|| strcmp(fname,"/dev/dsp1")==0){
			ret = remaudio_connect (fname,mode,NET_DSP);
		}else if (strcmp(fname,"/dev/mixer")==0){
			ret = remaudio_connect (fname,mode,NET_MIXER);
		}else if (strcmp(fname,"/dev/cdrom")==0){
			ret = remaudio_connect (fname,mode,NET_CDROM);
		}else if (strcmp(fname,"/dev/fd0")==0
			|| strcmp(fname,"/dev/fd0H1440")==0
			|| strcmp(fname,"/dev/fd1")==0
			|| strcmp(fname,"/dev/fd1H1440")==0){
			ret = remaudio_connect (fname,mode,NET_FLOPPY);
		}
	}else{
		if (strcmp(fname,"/dev/audio")==0
			|| strcmp(fname,"/dev/dsp")==0
			|| strcmp(fname,"/dev/dsp1")==0
			|| strcmp(fname,"/dev/mixer")==0
			|| strcmp(fname,"/dev/cdrom")==0
			|| strcmp(fname,"/dev/fd0")==0
			|| strcmp(fname,"/dev/fd0H1440")==0
			|| strcmp(fname,"/dev/fd1")==0
			|| strcmp(fname,"/dev/fd1H1440")==0){
			ret = __libc_open (fname,mode,perm);
		}
	}
	return ret;
}

static off_t remaudio_lseek(int fd, off_t offset, int whence)
{
	int ret = -1;
	REMAUDIO_TYPE type = tb[fd];
	if (type == NET_DSP
		|| type == NET_MIXER
		|| type == NET_AUDIO){
		errno = EINVAL;
		ret = -1;
	}else if (type == NET_FLOPPY || type == NET_CDROM){
		REMAUDIO_CTL ctl;
		ctl.cmd = NETCTL_LSEEK;
		ctl.p1 = offset;
		ctl.p2 = whence;
		remaudio_write (fd,ctl);
		ret = remaudio_readack(fd);
	}
	return ret;
}
static int remaudio_close (int fd)
{
	return __libc_close (fd);
}

static int remaudio_write (int fd, const void *buf, unsigned len)
{
	REMAUDIO_CTL ctl;
	ctl.cmd = NETCTL_WRITE;
	ctl.len = len;
	remaudio_write (fd,ctl);
	int ret = __libc_write (fd,buf,len);
	remaudio_readack(fd);
	return ret;
}

static int remaudio_netread (int fd, void *buf, unsigned len)
{
	REMAUDIO_CTL ctl;
	ctl.cmd = NETCTL_READ;
	ctl.len = len;
	int ret = -1;
	if (remaudio_write (fd,ctl)==sizeof(ctl)){
		n_debug ("ecrread ok\n");
		int n = __libc_read (fd,&ctl,sizeof(ctl));
		if (n==(int)sizeof(ctl)){
			int sofar = 0;
			int nb;
			n_debug ("essai de lire %d\n",ctl.len);
			while (sofar < ctl.len
				&& (nb=__libc_read(fd,((char*)buf)+sofar,len-sofar))>0){
				n_debug ("getting %d , sofar = %d, len = %d\n"
					,nb,sofar,ctl.len);
				sofar += nb;
			}
			n_debug ("obtient %d\n",sofar);
			ret = sofar;
			#if 0
				FILE *fau = fopen ("/tmp/log.au","a");
				fwrite (buf,1,ret,fau);
				fclose (fau);
			#endif
		}else{
			n_debug ("read ctl %d\n",n);
		}
	}else{
		n_debug ("ecrread kaput\n");
	}
	return ret;
}

static int remaudio_read (int fd, void *buf, unsigned len)
{
	REMAUDIO_TYPE type = NET_NOSPECIAL;
	if (fd >=0 && fd < MAX_FILE_TB) type = tb[fd];
	int ret = -1;
	if (type == NET_DSP){
		n_debug ("read from dsp %d\n",len);
		ret = remaudio_netread (fd,buf,len);
	}else if (type == NET_AUDIO){
		n_debug ("read from audio %d\n",len);
		ret = remaudio_netread (fd,buf,len);
	}else if (type == NET_MIXER){
		n_debug ("read from mixer %d\n",len);
		ret = remaudio_netread (fd,buf,len);
	}else if (type == NET_FLOPPY || type == NET_CDROM){
		n_debug ("read from floppy %d\n",len);
		ret = remaudio_netread (fd,buf,len);
	}
	return ret;
}

static struct {
	int request;
	const char *name;
	bool seen;
}tbreq[]={
	//{SNDCTL_PMGR_IFACE,			"SNDCTL_PMGR_IFACE",	false},
	{SNDCTL_DSP_RESET,			"SNDCTL_DSP_RESET",	false},
	{SNDCTL_DSP_SYNC,			"SNDCTL_DSP_SYNC",	false},
	{SNDCTL_DSP_SPEED,			"SNDCTL_DSP_SPEED",	false},
	{SNDCTL_DSP_STEREO,			"SNDCTL_DSP_STEREO",	false},
	{SNDCTL_DSP_GETBLKSIZE,		"SNDCTL_DSP_GETBLKSIZE",	false},
	{SNDCTL_DSP_SAMPLESIZE,		"SNDCTL_DSP_SAMPLESIZE",	false},
	{SNDCTL_DSP_CHANNELS,		"SNDCTL_DSP_CHANNELS",	false},
	{SOUND_PCM_WRITE_CHANNELS,	"SOUND_PCM_WRITE_CHANNELS",	false},
	{SOUND_PCM_WRITE_FILTER,	"SOUND_PCM_WRITE_FILTER",	false},
	{SNDCTL_DSP_POST,			"SNDCTL_DSP_POST",	false},
	{SNDCTL_DSP_SUBDIVIDE,		"SNDCTL_DSP_SUBDIVIDE",	false},
	{SNDCTL_DSP_SETFRAGMENT,	"SNDCTL_DSP_SETFRAGMENT",	false},
	{SNDCTL_DSP_GETFMTS,		"SNDCTL_DSP_GETFMTS",	false},
	{SNDCTL_DSP_SETFMT,			"SNDCTL_DSP_SETFMT",	false},
	{SNDCTL_DSP_GETOSPACE,		"SNDCTL_DSP_GETOSPACE",	false},
	{SNDCTL_DSP_GETISPACE,		"SNDCTL_DSP_GETISPACE",	false},
	{SNDCTL_DSP_NONBLOCK,		"SNDCTL_DSP_NONBLOCK",	false},
	{SNDCTL_DSP_GETCAPS,		"SNDCTL_DSP_GETCAPS",	false},
	{SNDCTL_DSP_GETTRIGGER,		"SNDCTL_DSP_GETTRIGGER",	false},
	{SNDCTL_DSP_SETTRIGGER,		"SNDCTL_DSP_SETTRIGGER",	false},
	{SNDCTL_DSP_GETIPTR,		"SNDCTL_DSP_GETIPTR",	false},
	{SNDCTL_DSP_GETOPTR,		"SNDCTL_DSP_GETOPTR",	false},
	{SNDCTL_DSP_MAPINBUF,		"SNDCTL_DSP_MAPINBUF",	false},
	{SNDCTL_DSP_MAPOUTBUF,		"SNDCTL_DSP_MAPOUTBUF",	false},
	{SNDCTL_DSP_SETSYNCRO,		"SNDCTL_DSP_SETSYNCRO",	false},
	{SNDCTL_DSP_SETDUPLEX,		"SNDCTL_DSP_SETDUPLEX",	false},
	{SOUND_PCM_READ_RATE,		"SOUND_PCM_READ_RATE",	false},
	{SOUND_PCM_READ_CHANNELS,	"SOUND_PCM_READ_CHANNELS",	false},
	{SOUND_PCM_READ_BITS,		"SOUND_PCM_READ_BITS",	false},
	{SOUND_PCM_READ_FILTER,		"SOUND_PCM_READ_FILTER",	false},

	{SNDCTL_SEQ_RESET,		"SNDCTL_SEQ_RESET",	false},
	{SNDCTL_SEQ_SYNC,		"SNDCTL_SEQ_SYNC",	false},
	{SNDCTL_SYNTH_INFO,		"SNDCTL_SYNTH_INFO",	false},
	{SNDCTL_SEQ_CTRLRATE,	"SNDCTL_SEQ_CTRLRATE",	false},
	{SNDCTL_SEQ_GETOUTCOUNT,	"SNDCTL_SEQ_GETOUTCOUNT",	false},
	{SNDCTL_SEQ_GETINCOUNT,	"SNDCTL_SEQ_GETINCOUNT",	false},
	{SNDCTL_SEQ_PERCMODE,	"SNDCTL_SEQ_PERCMODE",	false},
	{SNDCTL_FM_LOAD_INSTR,	"SNDCTL_FM_LOAD_INSTR",	false},
	{SNDCTL_SEQ_TESTMIDI,	"SNDCTL_SEQ_TESTMIDI",	false},
	{SNDCTL_SEQ_RESETSAMPLES,"SNDCTL_SEQ_RESETSAMPLES",	false},
	{SNDCTL_SEQ_NRSYNTHS,	"SNDCTL_SEQ_NRSYNTHS",	false},
	{SNDCTL_SEQ_NRMIDIS,		"SNDCTL_SEQ_NRMIDIS",	false},
	{SNDCTL_MIDI_INFO,		"SNDCTL_MIDI_INFO",	false},
	{SNDCTL_SEQ_THRESHOLD,	"SNDCTL_SEQ_THRESHOLD",	false},
	//{SNDCTL_SEQ_TRESHOLD,	"SNDCTL_SEQ_TRESHOLD",	false},
	{SNDCTL_SYNTH_MEMAVL,	"SNDCTL_SYNTH_MEMAVL",	false},
	{SNDCTL_FM_4OP_ENABLE,	"SNDCTL_FM_4OP_ENABLE",	false},
	//{SNDCTL_PMGR_ACCESS,		"SNDCTL_PMGR_ACCESS",	false},
	{SNDCTL_SEQ_PANIC,		"SNDCTL_SEQ_PANIC",	false},
	{SNDCTL_SEQ_OUTOFBAND,	"SNDCTL_SEQ_OUTOFBAND",	false},

	{SOUND_MIXER_READ_VOLUME,	"SOUND_MIXER_READ_VOLUME",false},
	{SOUND_MIXER_READ_BASS,		"SOUND_MIXER_READ_BASS",false},
	{SOUND_MIXER_READ_TREBLE,	"SOUND_MIXER_READ_TREBLE",false},
	{SOUND_MIXER_READ_SYNTH,	"SOUND_MIXER_READ_SYNTH",false},
	{SOUND_MIXER_READ_PCM,		"SOUND_MIXER_READ_PCM",false},
	{SOUND_MIXER_READ_SPEAKER,	"SOUND_MIXER_READ_SPEAKER",false},
	{SOUND_MIXER_READ_LINE,		"SOUND_MIXER_READ_LINE",false},
	{SOUND_MIXER_READ_MIC,		"SOUND_MIXER_READ_MIC",false},
	{SOUND_MIXER_READ_CD,		"SOUND_MIXER_READ_CD",false},
	{SOUND_MIXER_READ_IMIX,		"SOUND_MIXER_READ_IMIX",false},
	{SOUND_MIXER_READ_ALTPCM,	"SOUND_MIXER_READ_ALTPCM",false},
	{SOUND_MIXER_READ_RECLEV,	"SOUND_MIXER_READ_RECLEV",false},
	{SOUND_MIXER_READ_IGAIN,	"SOUND_MIXER_READ_IGAIN",false},
	{SOUND_MIXER_READ_OGAIN,	"SOUND_MIXER_READ_OGAIN",false},
	{SOUND_MIXER_READ_LINE1,	"SOUND_MIXER_READ_LINE1",false},
	{SOUND_MIXER_READ_LINE2,	"SOUND_MIXER_READ_LINE2",false},
	{SOUND_MIXER_READ_LINE3,	"SOUND_MIXER_READ_LINE3",false},

	{SOUND_MIXER_READ_RECSRC,	"SOUND_MIXER_READ_RECSRC",false},
	{SOUND_MIXER_READ_DEVMASK,	"SOUND_MIXER_READ_DEVMASK",false},
	{SOUND_MIXER_READ_RECMASK,	"SOUND_MIXER_READ_RECMASK",false},
	{SOUND_MIXER_READ_STEREODEVS,"SOUND_MIXER_READ_STEREODEVS",false},
	{SOUND_MIXER_READ_CAPS,		"SOUND_MIXER_READ_CAPS",false},


	{SOUND_MIXER_WRITE_VOLUME,	"SOUND_MIXER_WRITE_VOLUME",false},
	{SOUND_MIXER_WRITE_BASS,	"SOUND_MIXER_WRITE_BASS",false},
	{SOUND_MIXER_WRITE_TREBLE,	"SOUND_MIXER_WRITE_TREBLE",false},
	{SOUND_MIXER_WRITE_SYNTH,	"SOUND_MIXER_WRITE_SYNTH",false},
	{SOUND_MIXER_WRITE_PCM,		"SOUND_MIXER_WRITE_PCM",false},
	{SOUND_MIXER_WRITE_SPEAKER,	"SOUND_MIXER_WRITE_SPEAKER",false},
	{SOUND_MIXER_WRITE_LINE,	"SOUND_MIXER_WRITE_LINE",false},
	{SOUND_MIXER_WRITE_MIC,		"SOUND_MIXER_WRITE_MIC",false},
	{SOUND_MIXER_WRITE_CD,		"SOUND_MIXER_WRITE_CD",false},
	{SOUND_MIXER_WRITE_IMIX,	"SOUND_MIXER_WRITE_IMIX",false},
	{SOUND_MIXER_WRITE_ALTPCM,	"SOUND_MIXER_WRITE_ALTPCM",false},
	{SOUND_MIXER_WRITE_RECLEV,	"SOUND_MIXER_WRITE_RECLEV",false},
	{SOUND_MIXER_WRITE_IGAIN,	"SOUND_MIXER_WRITE_IGAIN",false},
	{SOUND_MIXER_WRITE_OGAIN,	"SOUND_MIXER_WRITE_OGAIN",false},
	{SOUND_MIXER_WRITE_LINE1,	"SOUND_MIXER_WRITE_LINE1",false},
	{SOUND_MIXER_WRITE_LINE2,	"SOUND_MIXER_WRITE_LINE2",false},
	{SOUND_MIXER_WRITE_LINE3,	"SOUND_MIXER_WRITE_LINE3",false},
	{SOUND_MIXER_INFO,			"SOUND_MIXER_INFO",false},
	{SOUND_MIXER_ACCESS,		"SOUND_MIXER_ACCESS",false},

	{-1,NULL,	false},
};



static void remaudio_identctl (int request, void *a1, void *a2, void *a3)
{
	int request_ffff = request & 0xffff;
	for (int i=0; tbreq[i].name != NULL; i++){
		//fprintf (stderr,"\t%x<>%x\n",request,tbreq[i].request);
		if ((tbreq[i].request & 0xffff) == request_ffff){
			if (!tbreq[i].seen){
				n_debug ("request %s %p %p %p\n",tbreq[i].name,a1,a2,a3);
				tbreq[i].seen = true;
			}
			break;
		}
	}
}


static int remaudio_ioctl (int fd, unsigned request, void *a1, void *a2, void *a3)
{
	REMAUDIO_TYPE type = NET_NOSPECIAL;
	if (fd >=0 && fd < MAX_FILE_TB) type = tb[fd];
	int ret = -1;
	if (type == NET_DSP){
		remaudio_identctl (request,a1,a2,a3);
		REMAUDIO_CTL ctl;
		ctl.cmd = NETCTL_IOCTL;
		ctl.request = request;
		ctl.p1 = 0;
		ctl.p2 = 0;
		ctl.p3 = 0;
		ret = 0;
		if (request == SNDCTL_DSP_SAMPLESIZE){
			int *p = (int*)a1;
			ctl.p1 = *p;
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
		}else if (request == SNDCTL_DSP_CHANNELS){
			ret = 1;
		}else if (request == SNDCTL_DSP_SPEED){
			int *p = (int*)a1;
			ctl.p1 = *p;
			printf ("speed = %d\n",*p);
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
		}else if (request == SNDCTL_DSP_RESET){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
		}else if (request == SNDCTL_DSP_GETCAPS){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}
	}else if (type == NET_AUDIO){
		n_debug ("ioctl %08x for audio\n",request);
		remaudio_identctl (request,a1,a2,a3);
	}else if (type == NET_MIXER){
		//fprintf (stderr,"ioctl %08x for mixer\n",request);
		REMAUDIO_CTL ctl;
		ctl.cmd = NETCTL_IOCTL;
		ctl.request = request;
		ctl.p1 = 0;
		ctl.p2 = 0;
		ctl.p3 = 0;
		if (request == SOUND_MIXER_READ_CAPS){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_DEVMASK){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_RECMASK){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_STEREODEVS){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_VOLUME
			|| request == SOUND_MIXER_READ_BASS
			|| request == SOUND_MIXER_READ_TREBLE
			|| request == SOUND_MIXER_READ_SYNTH
			|| request == SOUND_MIXER_READ_PCM
			|| request == SOUND_MIXER_READ_SPEAKER
			|| request == SOUND_MIXER_READ_LINE
			|| request == SOUND_MIXER_READ_MIC
			|| request == SOUND_MIXER_READ_CD
			|| request == SOUND_MIXER_READ_IMIX
			|| request == SOUND_MIXER_READ_ALTPCM
			|| request == SOUND_MIXER_READ_RECLEV
			|| request == SOUND_MIXER_READ_IGAIN
			|| request == SOUND_MIXER_READ_OGAIN
			|| request == SOUND_MIXER_READ_LINE1
			|| request == SOUND_MIXER_READ_LINE2
			|| request == SOUND_MIXER_READ_LINE3

			|| request == SOUND_MIXER_READ_RECSRC
			|| request == SOUND_MIXER_READ_DEVMASK
			|| request == SOUND_MIXER_READ_RECMASK
			|| request == SOUND_MIXER_READ_STEREODEVS
			|| request == SOUND_MIXER_READ_CAPS){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_WRITE_VOLUME
			|| request == SOUND_MIXER_WRITE_BASS
			|| request == SOUND_MIXER_WRITE_TREBLE
			|| request == SOUND_MIXER_WRITE_SYNTH
			|| request == SOUND_MIXER_WRITE_PCM
			|| request == SOUND_MIXER_WRITE_SPEAKER
			|| request == SOUND_MIXER_WRITE_LINE
			|| request == SOUND_MIXER_WRITE_MIC
			|| request == SOUND_MIXER_WRITE_CD
			|| request == SOUND_MIXER_WRITE_IMIX
			|| request == SOUND_MIXER_WRITE_ALTPCM
			|| request == SOUND_MIXER_WRITE_RECLEV
			|| request == SOUND_MIXER_WRITE_IGAIN
			|| request == SOUND_MIXER_WRITE_OGAIN
			|| request == SOUND_MIXER_WRITE_LINE1
			|| request == SOUND_MIXER_WRITE_LINE2
			|| request == SOUND_MIXER_WRITE_LINE3){
			int *p = (int*)a1;
			ctl.p1 = *p;
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else{
			fprintf (stderr,"unknown ioctl %08x for mixer\n",request);
		}
		remaudio_identctl (request,a1,a2,a3);
	}else if (type == NET_CDROM){
		REMAUDIO_CTL ctl;
		ctl.cmd = NETCTL_IOCTL;
		ctl.request = request;
		ctl.p1 = 0;
		ctl.p2 = 0;
		ctl.p3 = 0;
		//fprintf (stderr,"cdrom request %x\n",request);
		if (request == CDROMSUBCHNL){
			struct cdrom_subchnl *p = (struct cdrom_subchnl*)a1;
			remaudio_write (fd,ctl);
			__libc_write (fd,p,sizeof(*p));
			ret = remaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMREADTOCHDR){
			struct cdrom_tochdr *p = (struct cdrom_tochdr*)a1;
			remaudio_write (fd,ctl);
			__libc_write (fd,p,sizeof(*p));
			ret = remaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMREADTOCENTRY){
			struct cdrom_tocentry *p = (struct cdrom_tocentry*)a1;
			remaudio_write (fd,ctl);
			__libc_write (fd,p,sizeof(*p));
			ret = remaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMPLAYMSF){
			struct cdrom_msf *p = (struct cdrom_msf*)a1;
			remaudio_write (fd,ctl);
			__libc_write (fd,p,sizeof(*p));
			ret = remaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMVOLCTRL
			|| request == CDROMVOLREAD){
			struct cdrom_volctrl *p = (struct cdrom_volctrl*)a1;
			remaudio_write (fd,ctl);
			__libc_write (fd,p,sizeof(*p));
			ret = remaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMPAUSE
			|| request == CDROMRESUME
			|| request == CDROMSTOP
			|| request == CDROMSTART
			|| request == CDROMEJECT
			|| request == CDROMCLOSETRAY
			|| request == CDROMRESET){
			remaudio_write (fd,ctl);
			ret = remaudio_readack(fd,ctl);
		}else if (request == CDROMPLAYTRKIND
			|| request == CDROMREADMODE2
			|| request == CDROMREADMODE1
			|| request == CDROMREADAUDIO
			|| request == CDROMEJECT_SW
			|| request == CDROMMULTISESSION
			|| request == CDROM_GET_UPC
			|| request == CDROMVOLREAD){
			#if 0
				int *p = (int*)a1;
				ctl.p1 = *p;
				remaudio_write (fd,ctl);
				ret = remaudio_readack(fd,ctl);
				*(int*)a1 = ctl.p1;
			#else
				printf ("CDROM ioctl %08x pas fait\n",request);
			#endif
		}else{
			fprintf (stderr,"unknown ioctl %08x for cdrom\n",request);
		}
	}else if (type == NET_FLOPPY){
		fprintf (stderr,"unknown ioctl %08x for floppy\n",request);
		errno = EINVAL;
		ret = -1;
	}
	return ret;
}

static DISPATCH m={
	NULL,
	remaudio_open,
	remaudio_close,
	remaudio_write,
	remaudio_read,
	remaudio_ioctl,
	remaudio_lseek,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,		// socket
	NULL,		// connect
	NULL,		// bind
	NULL,		// chmod
	NULL,		// chown
	NULL,		// lchown
	NULL,		// readlink
	NULL,		// utimes
	NULL,		// access
	NULL,		// euidaccess
	NULL,		// link
	NULL,		// symlink
	NULL,		// fchmod
	NULL,		// fchown
	NULL,		// execve
};

static REGISTER_DISPATCH register_m (m);
